"""The almighty Hello World! example"""
# We'll use sys to properly exit with an error code.
import os
import sys
import sdl2.ext

# Create a resource container, so that we can easily access all
# resource, we bundle with our application. We are using the current
# file's location and define the "resources" subdirectory as the
# location, in which we keep all data.
filepath = os.path.abspath(os.path.dirname(__file__))
RESOURCES = sdl2.ext.Resources(filepath, "resources")


def run():
    # Initialize the video system - this implicitly initializes some
    # necessary parts within the SDL2 DLL used by the video module.
    #
    # You SHOULD call this before using any video related methods or
    # classes.
    sdl2.ext.init()

    # Create a new window (like your browser window or editor window,
    # etc.) and give it a meaningful title and size. We definitely need
    # this, if we want to present something to the user.
    window = sdl2.ext.Window("Hello World!", size=(592, 460))

    # By default, every Window is hidden, not shown on the screen right
    # after creation. Thus we need to tell it to be shown now.
    window.show()

    # Create a Renderer for the new window, which we can use to copy and
    # draw things to the screen. Renderers can use hardware-accelerated
    # backends (e.g. OpenGL, Direct3D) as well as software-accelerated ones,
    # depending on the flags you create it with.
    renderflags = sdl2.SDL_RENDERER_SOFTWARE
    if "-hardware" in sys.argv:
        renderflags = (
            sdl2.SDL_RENDERER_ACCELERATED | sdl2.SDL_RENDERER_PRESENTVSYNC
        )
    renderer = sdl2.ext.Renderer(window, flags=renderflags)

    # Import an image file and convert it to a Texture. A Texture is an SDL
    # surface that has been prepared for use with a given Renderer. 
    tst_img = sdl2.ext.load_bmp(RESOURCES.get_path("hello.bmp"))
    tx = sdl2.ext.Texture(renderer, tst_img)

    # Display the image on the window. This code takes the texture we created
    # earlier and copies it to the renderer (with the top-left corner of the
    # texture placed at the coordinates (0, 0) on the window surface), then 
    # takes the contents of the renderer surface and presents them on its
    # associated window.
    renderer.copy(tx, dstrect=(0, 0))
    renderer.present()

    # Create a simple event loop. This fetches the SDL2 event queue and checks
    # for any quit events. Once a quit event is received, the loop will end
    # and we'll send the signal to quit the program.
    running = True
    while running:
        events = sdl2.ext.get_events()
        for event in events:
            if event.type == sdl2.SDL_QUIT:
                running = False
                break

    # Now that we're done with the SDL2 library, we shut it down nicely using
    # the `sdl2.ext.quit` function.
    sdl2.ext.quit()
    return 0

if __name__ == "__main__":
    sys.exit(run())



import sys
import pytest
from ctypes import c_char_p, c_void_p, cast
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_QuitSubSystem, SDL_INIT_EVERYTHING
from sdl2 import events

# NOTE: These tests are very incomplete and in need of a rewrite


# Test initializing event structs (is this actually useful?)

def test_SDL_AudioDeviceEvent():
    event = events.SDL_AudioDeviceEvent()
    assert isinstance(event, events.SDL_AudioDeviceEvent)

def test_SDL_DisplayEvent():
    event = events.SDL_DisplayEvent()
    assert isinstance(event, events.SDL_DisplayEvent)

def test_SDL_WindowEvent():
    event = events.SDL_WindowEvent()
    assert isinstance(event, events.SDL_WindowEvent)

def test_SDL_KeyboardEvent():
    event = events.SDL_KeyboardEvent()
    assert isinstance(event, events.SDL_KeyboardEvent)

def test_SDL_TextEditingEvent():
    event = events.SDL_TextEditingEvent()
    assert isinstance(event, events.SDL_TextEditingEvent)

def test_SDL_TextInputEvent():
    event = events.SDL_TextInputEvent()
    assert isinstance(event, events.SDL_TextInputEvent)

def test_SDL_MouseMotionEvent():
    event = events.SDL_MouseMotionEvent()
    assert isinstance(event, events.SDL_MouseMotionEvent)

def test_SDL_MouseButtonEvent():
    event = events.SDL_MouseButtonEvent()
    assert isinstance(event, events.SDL_MouseButtonEvent)

def test_SDL_MouseWheelEvent():
    event = events.SDL_MouseWheelEvent()
    assert isinstance(event, events.SDL_MouseWheelEvent)

def test_SDL_JoyAxisEvent():
    event = events.SDL_JoyAxisEvent()
    assert isinstance(event, events.SDL_JoyAxisEvent)

def test_SDL_JoyBallEvent():
    event = events.SDL_JoyBallEvent()
    assert isinstance(event, events.SDL_JoyBallEvent)

def test_SDL_JoyHatEvent():
    event = events.SDL_JoyHatEvent()
    assert isinstance(event, events.SDL_JoyHatEvent)

def test_SDL_JoyButtonEvent():
    event = events.SDL_JoyButtonEvent()
    assert isinstance(event, events.SDL_JoyButtonEvent)

def test_SDL_JoyDeviceEvent():
    event = events.SDL_JoyDeviceEvent()
    assert isinstance(event, events.SDL_JoyDeviceEvent)

def test_SDL_ControllerAxisEvent():
    event = events.SDL_ControllerAxisEvent()
    assert isinstance(event, events.SDL_ControllerAxisEvent)

def test_SDL_ControllerButtonEvent():
    event = events.SDL_ControllerButtonEvent()
    assert isinstance(event, events.SDL_ControllerButtonEvent)

def test_SDL_ControllerDeviceEvent():
    event = events.SDL_ControllerDeviceEvent()
    assert isinstance(event, events.SDL_ControllerDeviceEvent)

def test_SDL_ControllerTouchpadEvent():
    event = events.SDL_ControllerTouchpadEvent()
    assert isinstance(event, events.SDL_ControllerTouchpadEvent)

def test_SDL_ControllerSensorEvent():
    event = events.SDL_ControllerSensorEvent()
    assert isinstance(event, events.SDL_ControllerSensorEvent)

def test_SDL_TouchFingerEvent():
    event = events.SDL_TouchFingerEvent()
    assert isinstance(event, events.SDL_TouchFingerEvent)

def test_SDL_MultiGestureEvent():
    event = events.SDL_MultiGestureEvent()
    assert isinstance(event, events.SDL_MultiGestureEvent)

def test_SDL_DollarGestureEvent():
    event = events.SDL_DollarGestureEvent()
    assert isinstance(event, events.SDL_DollarGestureEvent)

def test_SDL_DropEvent():
    event = events.SDL_DropEvent()
    assert isinstance(event, events.SDL_DropEvent)

def test_SDL_SensorEvent():
    event = events.SDL_SensorEvent()
    assert isinstance(event, events.SDL_SensorEvent)

def test_SDL_QuitEvent():
    event = events.SDL_QuitEvent()
    assert isinstance(event, events.SDL_QuitEvent)

def test_SDL_UserEvent():
    event = events.SDL_UserEvent()
    assert isinstance(event, events.SDL_UserEvent)

def test_SDL_SysWMEvent():
    event = events.SDL_SysWMEvent()
    assert isinstance(event, events.SDL_SysWMEvent)

def test_SDL_Event():
    event = events.SDL_Event()
    assert isinstance(event, events.SDL_Event)


class TestSDLEvents(object):
    __tags__ = ["sdl"]

    @classmethod
    def setup_class(cls):
        SDL_Init(SDL_INIT_EVERYTHING)

    @classmethod
    def teardown_class(cls):
        SDL_QuitSubSystem(SDL_INIT_EVERYTHING)
        SDL_Quit()

    def test_SDL_AddDelEventWatch(self):
        eventwatch = []

        def watch(data, event):
            eventwatch.append((event.contents, data,))
            return 0
        efilter = events.SDL_EventFilter(watch)
        udata = c_char_p(b"Something random")
        events.SDL_AddEventWatch(efilter, cast(udata, c_void_p))
        ev = events.SDL_Event()
        ev.type = events.SDL_USEREVENT
        ev.user = events.SDL_UserEvent()
        events.SDL_PushEvent(ev)
        assert len(eventwatch) == 1
        # TODO: x
        # self.assertEqual(eventwatch[0][1], udata)

        events.SDL_DelEventWatch(efilter, udata)
        ev = events.SDL_Event()
        events.SDL_PushEvent(ev)
        assert len(eventwatch) == 1
        # TODO: x
        # self.assertEqual(eventwatch[0][1], udata)

    @pytest.mark.skip("not implemented")
    def test_SDL_EventState(self):
        pass
        # state = events.SDL_EventState(events.SDL_USEREVENT, events.SDL_QUERY)
        # self.assertEqual(state, events.SDL_ENABLE)
        # state = events.SDL_EventState(events.SDL_USEREVENT,events.SDL_IGNORE)
        # self.assertEqual(state, events.SDL_ENABLE)
        # state = events.SDL_EventState(events.SDL_USEREVENT, events.SDL_QUERY)
        # self.assertEqual(state, events.SDL_IGNORE)
        # state = events.SDL_EventState(events.SDL_USEREVENT,events.SDL_ENABLE)
        # self.assertEqual(state, events.SDL_IGNORE)
        # state = events.SDL_EventState(events.SDL_USEREVENT, events.SDL_QUERY)
        # self.assertEqual(state, events.SDL_ENABLE)

        # self.assertRaises(TypeError, events.SDL_EventState, None, None)

        # ev = events.SDL_Event()
        # ev.type = events.SDL_USEREVENT
        # ev.user = events.SDL_UserEvent()
        # events.SDL_PushEvent(ev)

    @pytest.mark.skip("not implemented")
    def test_SDL_GetEventState(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_FilterEvents(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_FlushEvent(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_FlushEvents(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_GetSetEventFilter(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_HasEvent(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_HasEvents(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_PeepEvents(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_PollEvent(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_PumpEvents(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_PushEvent(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_RegisterEvents(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_WaitEvent(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_WaitEventTimeout(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_SDL_QuitRequested(self):
        pass



import sys
import pytest
import sdl2
from sdl2.stdinc import SDL_TRUE, SDL_FALSE
from sdl2.error import SDL_GetError, SDL_ClearError
from sdl2 import hidapi

# Make sure hidapi subsystem is available and works before running tests
skipmsg = 'HIDAPI requires SDL 2.0.18 or newer'
pytestmark = pytest.mark.skipif(sdl2.dll.version < 2018, reason=skipmsg)


@pytest.fixture
def hidapi_setup():
    SDL_ClearError()
    ret = hidapi.SDL_hid_init()
    assert ret == 0
    yield ret  # Run the test
    ret = hidapi.SDL_hid_exit()
    assert ret == 0


# NOTE: Remove this xfail once libudev is officially removed from pysdl2-dll
@pytest.mark.xfail("linux" in sys.platform, reason="udev problems")
def test_SDL_hid_init_exit():
    SDL_ClearError()
    # Initialize the library
    ret = hidapi.SDL_hid_init()
    assert SDL_GetError() == b""
    assert ret == 0
    # Exit the library
    ret = hidapi.SDL_hid_exit()
    assert SDL_GetError() == b""
    assert ret == 0


def test_SDL_hid_device_change_count(hidapi_setup):
    # Check for HID device changes
    ret = hidapi.SDL_hid_device_change_count()
    assert ret >= 0


def test_SDL_hid_enumerate(hidapi_setup):
    devices = hidapi.SDL_hid_enumerate(0, 0)
    assert SDL_GetError() == b""
    if devices != None:
        hidapi.SDL_hid_free_enumeration(devices)
        assert SDL_GetError() == b""


@pytest.mark.skip("not implemented")
def test_SDL_hid_open_close():
    # SDL_hid_open & SDL_hid_close
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_open_path():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_write():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_read_timeout():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_read():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_set_nonblocking():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_send_feature_report():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_get_feature_report():
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_get_info_strings():
    # SDL_hid_get_manufacturer_string & SDL_hid_get_product_string &
    # SDL_hid_get_serial_number_string & SDL_hid_get_indexed_string
    pass

@pytest.mark.skip("not implemented")
def test_SDL_hid_ble_scan():
    pass



import sys
import pytest
from sdl2 import ext as sdl2ext

# NOTE: Deprecated, so I'm not writing tests for it


class TestExtUIFactory(object):
    __tags__ = ["sdl", "sdl2ext"]

    @pytest.mark.skip("not implemented")
    def test_init(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_create_button(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_create_checkbutton(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_create_text_entry(self):
        pass


class TestExtUIProcessor(object):
    __tags__ = ["sdl", "sdl2ext"]

    @pytest.mark.skip("not implemented")
    def test_init(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_activate(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_deactivate(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_dispatch(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_mousedown(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_mouseup(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_mousemotion(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_passevent(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_process(self):
        pass

    @pytest.mark.skip("not implemented")
    def test_textinput(self):
        pass



import os
import re
from ctypes import c_int, byref
from .. import surface, pixels, rwops, rect
from .compat import byteify, stringify, utf8, _is_text
from .common import raise_sdl_err
from .color import Color, convert_to_color
from .draw import prepare_color
from .resources import _validate_path
from .surface import _create_surface

_HASSDLTTF = True
try:
    from .. import sdlttf
except ImportError:
    _HASSDLTTF = False

__all__ = ["FontTTF", "FontManager"]


# This module variable keeps track of the number of separate times SDL_ttf has
# been initialized. This is important to keep track of, since de-initializing
# the library frees all associated font data, so we want to make sure this
# hasn't changed during the lifetime of an object to avoid segfaults.
_ttf_init_count = 0


def _ttf_init():
    global _ttf_init_count
    if not _HASSDLTTF:
        raise RuntimeError("SDL_ttf is required, but is not installed.")

    # Check if TTF already initialized, return immediately if it was
    if sdlttf.TTF_WasInit() > 0:
        return _ttf_init_count

    # Handle a weirdness in how TTF_Init and TTF_Quit work: TTF_Init
    # blindly increments TTF_WasInit every time it's called and TTF_Quit
    # blindly decrements it, but TTF_Quit only *actually* quits when 
    # TTF_WasInit - 1 == 0. Here, we try to ensure we're starting at 0.
    while sdlttf.TTF_WasInit() < 1:
        ret = sdlttf.TTF_Init()
        if ret != 0:
            raise_sdl_err("initializing the SDL_ttf library")

    # If initialized successfully, update and return the global init count
    _ttf_init_count += 1
    return _ttf_init_count


def _ttf_quit():
    if not _HASSDLTTF:
        raise RuntimeError("SDL_ttf is required, but is not installed.")

    # Make sure WasInit is non-negative before trying to quit
    while sdlttf.TTF_WasInit() < 1:
        ret = sdlttf.TTF_Init()
        if ret != 0:
            raise_sdl_err("initializing the SDL_ttf library")

    # Actually quit the library (won't really quit until TTF_WasInit == 0)
    while sdlttf.TTF_WasInit > 0:
        sdlttf.TTF_Quit()


def _is_whitespace(s):
    ws_chars = [" ", "\n", "\t", "\r"]
    return len(s) == 0 or all([c in ws_chars for c in s])


def _split_on_whitespace(line):
    # Splits a string on whitespace starts without removing the whitespace
    words = []
    current_word = ""
    for s in re.split(r'(\s+)', line):
        current_word += s
        if not _is_whitespace(s):
            words.append(current_word)
            current_word = ""
    return words



class FontTTF(object):
    """A class for rendering text with a given TrueType font.

    This class loads a TrueType or OpenType font using the **SDL_ttf** library
    and allows the user to render text with it in various sizes and colors.

    To simplify rendering text with different font sizes and colors, this class
    allows users to define font styles (e.g. 'instructions', 'error', 'title',
    etc.) using :meth:`add_style`, which can then be used by name with the
    :meth:`render_text` method.

    By default, the :class:`FontTTF` class defines font sizes in points (pt), a
    common unit of font size used by many libraries and text editors. However,
    you can also define font sizes in units of *pixel height* (px), i.e. the
    ascent height of the font's tallest alphanumeric ASCII character. This is
    done by calculating each font's px-to-pt scaling factor on import, meaning
    that some rounding error may occur. If using a font that does not include
    all alphanumeric ASCII characters (A-Z, a-z, 0-9) or wanting to scale the
    maximum pixel height to a different subset of characters (e.g. just 0-9
    digits), you will need to specifiy a custom 'height_chars' string when
    creating the font object.

    .. note::
       If loading a font from an SDL RWops file object, you must not free the
       file object until you are done with the font. Otherwise, SDL_ttf will
       try to render with a closed font and hard-crash Python. 
    
    Args:
        font (str or :obj:`SDL_RWops`): The relative (or absolute) path to the
            font to load, or an SDL file object containing the font data.
        size (int or str): The default size for the font, either as an integer
            (assumed to be in pt) or as a string specifying the unit of size
            (e.g. '12pt' or '22px').
        color (~sdl2.ext.Color or tuple): The default color to use for
            rendering text. Defaults to white if not specified.
        index (int, optional): The index of the font face to load if the font
            file contains multiple faces. Defaults to 0 (first face in the file)
            if not specified.
        height_chars (str, optional): The set of font characters to use for
            calculating the maximum height (in pixels) of the font for 
    
    """
    def __init__(self, font, size, color, index=0, height_chars=None):
        # Load the font, either from a file or an SDL file object
        if isinstance(font, rwops.SDL_RWops):
            fname = None
            self._font_file = None
            self._font_rw = font
        elif _is_text(font):
            fullpath, fname = _validate_path(font, "a font")
            fullpath = byteify(fullpath)
            self._font_file = open(fullpath, "rb")
            self._font_rw = rwops.rw_from_object(self._font_file)
        else:
            e = "'font' must be a path string or an SDL file object (got {0})."
            raise ValueError(e.format(type(font).__str__))
    
        # Initialize the TTF library (if not already initialized)
        self._ttf_init_count = _ttf_init()
        self._index = index

        # Get the px-to-pt scaling factor for the font
        if not height_chars:
            caps = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            height_chars = caps + caps.lower() + "0123456789"
        self._scale_factor = self._get_scale_factor(height_chars, fname)

        # Initialize font styles and add the default style    
        self._styles = {}
        self.add_style("default", size, color)

    def _check_if_closed(self):
        # Makes sure the font hasn't been closed, raising an exception if it has
        ttf_was_closed = _ttf_init_count > self._ttf_init_count
        if self._font_rw == None:
            e = "The font has been closed and can no longer be used."
            raise RuntimeError(e)
        elif sdlttf.TTF_WasInit() < 1 or ttf_was_closed:
            e = "The SDL_ttf library has been de-initialized since creating the "
            e += "font. The font can no longer be used."
            raise RuntimeError(e)

    def _get_scale_factor(self, testchars, fname):
        # Calculate the pt -> px conversion factor for the current font.
        tmpsize = 40  # Font size in pt at which to calculate pt -> px scale factor
        max_ascent = 0
        minX, maxX, minY, maxY = c_int(0), c_int(0), c_int(0), c_int(0)
        advance = c_int(0)
        tmpfont = sdlttf.TTF_OpenFontIndexRW(self._font_rw, 0, tmpsize, self._index)
        if not tmpfont:
            e = "initializing the font"
            if fname:
                e += " '{0}'".format(fname)
            raise_sdl_err(e)
        for char in testchars:
            if sdlttf.TTF_GlyphIsProvided(tmpfont, ord(char)) == 0:
                e = "The current font does not provide a glyph for character "
                e += "'{0}'. Please initialize the font with a different "
                e += "'height_chars' string."
                sdlttf.TTF_CloseFont(tmpfont)
                raise RuntimeError(e.format(char.encode('utf-8')))
            sdlttf.TTF_GlyphMetrics(
                tmpfont, ord(char), byref(minX), byref(maxX),
                byref(minY), byref(maxY), byref(advance)
            )
            if maxY.value > max_ascent:
                max_ascent = maxY.value
        sdlttf.TTF_CloseFont(tmpfont)
        return tmpsize / float(max_ascent)

    def _parse_size(self, size):
        # Parse and validate a font size, which can be an int or a string
        if isinstance(size, str):
            if '-' in size or '.' in size:
                e = "Font size must be a positive whole number (got '{0}')."
                raise ValueError(e.format(size))
            sz = "".join([i for i in size if i.isdigit()])
            unit = size[len(sz):]
            sz = int(sz)
            if unit == "px":
                sz = int(sz * self._scale_factor)
            elif unit not in ["pt", "px"]:
                e = "Font size units must be either 'pt' or 'px' (got '{0}')."
                raise ValueError(e.format(unit))
        else:
            sz = int(size)
        return sz

    def _parse_line_height(self, height, default_skip):
        if height is None:
            h = default_skip
        elif isinstance(height, str):
            # If line height specified as a percentage of the default line skip
            if height[-1] == "%":
                pct = float(height[:-1])
                if not pct > 0:
                    e = "Relative line heights must be greater than 0% (got '{0}')."
                    raise ValueError(e.format(height))
                h = int(default_skip * (pct / 100.0))
            # If line height specified explicitly in pixels
            elif height[-2:] == "px":
                if '-' in height or '.' in height:
                    e = "Line height must be a positive whole number (got '{0}')."
                    raise ValueError(e.format(height))
                h = int("".join([i for i in height if i.isdigit()]))
            else:
                e = "Line height units must be either 'px' or '%' (got '{0}')."
                raise ValueError(e.format(height))
        else:
            # If numeric, assume pixels
            if height != int(height) or height < 1:
                e = "Line height must be a positive integer (got {0})."
                raise ValueError(e.format(height))
            h = int(height)
        return max(h, 1)

    def _get_line_size(self, line, style):
        # Get the height and width of a given line of text in a given style
        font = self._styles[style]['font']
        if len(line) == 0:
            return (0, sdlttf.TTF_FontHeight(font))
        lw, lh = c_int(0), c_int(0)
        ret = sdlttf.TTF_SizeUTF8(font, byteify(line), byref(lw), byref(lh))
        if ret != 0:
            raise_sdl_err("rendering text with the '{0}' style".format(style))
        return (lw.value, lh.value)

    def _wrap_line(self, line, style, line_width, width):
        # Splits a single line of text into separate lines in order to fit
        # within a given width with a given text style
        lines = []
        while line_width > width:

            # First, estimate the position of the wrap word and move to the
            # end of it
            pos = int((width / float(line_width)) * len(line))
            words = _split_on_whitespace(line)
            for i in range(len(words)):
                split_word = i + 1
                segment = "".join(words[:split_word])
                if len(segment) >= pos and len(segment) < len(line):
                    pos = len(segment)
                    break

            # Then, work backwards until the line fits within the width
            segment_w = self._get_line_size(segment, style)[0]
            while segment_w > width:
                if split_word > 1:
                    split_word -= 1
                    segment = "".join(words[:split_word])
                    pos = len(segment)
                elif pos > 1:
                    pos -= 1
                    segment = segment[:pos]
                else:
                    e = "Character '{0}' exceeds maximum width of text"
                    raise RuntimeError(e.format(line[0]))
                segment_w = self._get_line_size(segment, style)[0]

            # Append the trimmed line & remove the text from the line, then
            # recalculate the width of the remaining text to see if it fits
            lines.append(segment)
            line = line[pos:].lstrip()
            line_width = self._get_line_size(line, style)[0]

        lines.append(line)
        return lines

    def _split_lines(self, text, style, width=None):
        # Splits one or more lines of text to ensure they all fit within a given
        # width (in pixels) when rendered with a given text style
        unwrapped_lines = text.split("\n")
        if width is None:
            return unwrapped_lines

        lines = []
        for line in unwrapped_lines:
            lw = self._get_line_size(line, style)[0]
            if lw > width:
                # If line is too long and contains trailing whitespace, try w/o
                if len(line) < len(line.rstrip()):
                    line = line.rstrip()
                    lw = self._get_line_size(line, style)[0]
                # If line width exceeds wrap width, we break it into smaller lines
                lines += self._wrap_line(line, style, lw, width)
            else:
                lines.append(line)

        return lines

    def _render_lines(self, lines, style, line_h, width, align):
        # Renders one or more lines of text to an SDL surface, optionally with
        # a custom line height, wrap width, and/or text alignment
        font = self._styles[style]['font']
        color = self._styles[style]['color']
        bg_col = self._styles[style]['bg']
        if not bg_col:
            c = color
            bg_col = (c.r, c.g, c.b, 0)  # Default bg = font color w/ full transparency

        # Actually render the text
        rendered = []
        for line in lines:
            if len(line) == 0:
                rendered.append(None)
                continue
            fontsf = sdlttf.TTF_RenderUTF8_Blended(font, byteify(line), color)
            if not fontsf:
                raise_sdl_err("rendering text with the '{0}' style".format(style))
            rendered.append(fontsf)

        # If single line of text w/ no background fill or custom width, return as-is
        if len(rendered) == 1 and width == None and bg_col is None:
            return rendered[0]

        # Determine height and width of background surface
        font_height = sdlttf.TTF_FontHeight(font)
        default_skip = sdlttf.TTF_FontLineSkip(font)
        line_h = self._parse_line_height(line_h, default_skip)
        height = line_h * (len(lines) - 1) + font_height
        if width == None:
            width = max([line.contents.w for line in rendered])

        # Create background surface for the text and render lines to it
        sf = _create_surface((width, height), bg_col, errname="background")
        line_y = 0
        for line in rendered:
            if line:
                lw, lh = (line.contents.w, line.contents.h)
                if align == "left":
                    line_x = 0
                elif align == "center":
                    line_x = int((width - lw) / 2)
                elif align == "right":
                    line_x = width - lw
                line_rect = rect.SDL_Rect(line_x, line_y, lw, lh)
                surface.SDL_BlitSurface(line, None, sf, line_rect)
                surface.SDL_FreeSurface(line)
            line_y += line_h

        return sf

    def get_ttf_font(self, style='default'):
        """Returns the base :obj:`~sdl2.sdlttf.TTF_Font` for a given font style.

        This method provides access to the base ctypes object for each font
        style so that they can be used with the full set of :mod:`~sdl2.sdlttf`
        functions.

        Args:
            style (str, optional): The font style for which the base ctypes font
                object will be retrieved. Defaults to the 'default' style if not
                specified.

        Returns:
            :obj:`~sdl2.sdlttf.TTF_Font`: The base ctypes font for the style.

        """
        self._check_if_closed()
        if style not in self._styles.keys():
            e = "The '{0}' style is not defined for the current font."
            raise ValueError(e.format(style))
        return self._styles[style]['font'].contents

    def add_style(self, name, size, color, bg_color=None):
        """Defines a named font style for the current font.

        Currently, a font style defines a font size, color, and background color
        combination to use for rendering text. Additional style attributes may
        be added in future releases.

        Args:
            name (str): The name of the new font style (e.g. 'title').
            size (int or str): The font size for the style, either as an integer
                (assumed to be in pt) or as a string specifying the unit of size
                (e.g. '12pt' or '22px').
            color (~sdl2.ext.Color): The font color for the style.
            bg_color (~sdl2.ext.Color, optional): The background surface color
                for the style. Defaults to a fully-transparent background if not
                specified.
    
        """
        # Preprocess and validate style values
        self._check_if_closed()
        if name in self._styles.keys():
            e = "A style named '{0}' is already defined for the current font."
            raise ValueError(e.format(name))
        size_pt = self._parse_size(size)
        if size_pt < 1:
            e = "Font size must be at least 1pt, got {0} (specified as '{1}')"
            raise ValueError(e.format(size_pt, size))
        c = convert_to_color(color)
        if bg_color is not None:
            bg_color = convert_to_color(bg_color)

        # Actually create font object for style
        rwops.SDL_RWseek(self._font_rw, 0, rwops.RW_SEEK_SET)
        font = sdlttf.TTF_OpenFontIndexRW(self._font_rw, 0, size_pt, self._index)
        if not font:
            raise_sdl_err("initializing the '{0}' font style".format(name))
        
        self._styles[name] = {
            'font': font,
            'size': size_pt,
            'color': pixels.SDL_Color(c.r, c.g, c.b, c.a),
            'bg': bg_color
        }

    def render_text(self, text, style='default', line_h=None, width=None, align='left'):
        """Renders a string of text to a new surface.

        If a newline character (``\\n``) is encountered in the string, it will
        be rendered as a line break in the rendered text. Additionally, if a
        surface width is specified, any lines of text that exceed the width of
        the surface will be wrapped. Multi-line text can be left-aligned
        (the default), right-aligned, or centered, and the spacing between lines
        can be modified using the `line_h` argument.

        Line heights can be specified in pixels (e.g. ``20`` or ``'20px'``) or
        as percentages of the TTF-suggested line spacing for the font (e.g.
        ``'150%'``).

        Args:
            text (str): The string of text to render to the target surface.
            style (str, optional): The font style with which to render the
                given string. Defaults to the 'default' style if not specified.
            line_h (int or str, optional): The line height to use for each
                line of the rendered text, either in pixels or as a percentage
                of the font's suggested line height. If not specified, the
                suggested line height for the font will be used.
            width (int, optional): The width (in pixels) of the output surface.
                If a line of text exceeds this value, it will be automatically
                wrapped to fit within the specified width. Defaults to ``None``.
            align (str, optional): The alignment of lines of text for multi-line
                strings. Can be 'left' (left-aligned), 'right' (right-aligned),
                or 'center' (centered). Defaults to 'left'.

        Returns:
            :obj:`~sdl2.SDL_Surface`: A 32-bit ARGB surface containing the
            rendered text.

        """
        self._check_if_closed()
        if len(text) == 0:
            raise ValueError("Cannot render an empty string.")
        if style not in self._styles.keys():
            e = "The '{0}' style is not defined for the current font."
            raise ValueError(e.format(style))
        if align not in ["left", "right", "center"]:
            e = "Text alignment mode must be 'left', 'right', or 'center'."
            raise ValueError(e)

        # Actually render the text
        lines = self._split_lines(utf8(text), style, width)
        sf = self._render_lines(lines, style, line_h, width, align)
        return sf.contents

    def close(self):
        """Cleanly closes the font and frees all associated memory.

        This method should be called when you are finished with the font to free
        the resources taken up by the font and its styles. Once this method has
        been called, the font can no longer be used.

        """
        ttf_was_closed = _ttf_init_count > self._ttf_init_count
        if self._font_rw != None:
            if sdlttf.TTF_WasInit() > 0 and not ttf_was_closed:
                for name, style in self._styles.items():
                    sdlttf.TTF_CloseFont(style['font'])
            self._styles = None
            if self._font_file:
                rwops.SDL_RWclose(self._font_rw)
            self._font_rw = None
            self._font_file = None

    def contains(self, c):
        """Checks whether a given character exists within the font.
        
        Args:
            c (str): The glpyh (i.e. character) to check for within the font.

        Returns:
            bool: ``True`` if the font contains the glpyh, otherwise ``False``.
    
        """
        self._check_if_closed()
        font = self._styles['default']['font']
        return sdlttf.TTF_GlyphIsProvided(font, ord(c)) != 0

    @property
    def family_name(self):
        """str: The family name (e.g. "Helvetica") of the loaded font.
        
        """
        self._check_if_closed()
        name = sdlttf.TTF_FontFaceFamilyName(self._styles['default']['font'])
        return stringify(name) if name else None

    @property
    def style_name(self):
        """str: The style name (e.g. "Bold") of the loaded font.
        
        """
        self._check_if_closed()
        name = sdlttf.TTF_FontFaceStyleName(self._styles['default']['font'])
        return stringify(name) if name else None

    @property
    def is_fixed_width(self):
        """bool: Whether the current font face is fixed-width (i.e. monospaced).

        If ``True``, all characters in the current font have the same width.

        """
        self._check_if_closed()
        font = self._styles['default']['font']
        return sdlttf.TTF_FontFaceIsFixedWidth(font) != 0



class FontManager(object):
    """A class for managing and rendering TrueType fonts.

    .. note:: 
       This class is has been deprecated in favor of the more flexible
       :class:`~sdl2.ext.FontTTF` class.
    
    This class provides a basic wrapper around the SDL_ttf library. One font
    path must be given to initialise the FontManager.
    
    The first face is always at index 0. It can be used for TTC (TrueType Font
    Collection) fonts.

    Args:
        font_path (str): The relative (or absolute) path to the font
            to load.
        alias (str, optional): The name to give the font within the
            FontManager. Defaults to the font filename if not specified.
        size (int, optional): The size (in pt) at which to load the default
            font. Defaults to 16pt if not specified.
        color (~sdl2.ext.Color): The default font rendering color. Defaults
            to opaque white if not specified.
        bg_color (~sdl2.ext.Color, optional): The default background surface
            color. Defaults to a fully-transparent background if not specified.
        index (int, optional): The index of the font face to load if the
            font file contains multiple faces. Defaults to 0 (first face in
            the file) if not specified.

    Attributes:
        size (int): The default font size in pt.

    """
    def __init__(self, font_path, alias=None, size=16,
                 color=Color(255, 255, 255), bg_color=Color(0, 0, 0), index=0):
        self._ttf_init_count = 0
        self._ttf_init_count = _ttf_init()
        self.fonts = {}  # fonts = {alias: {size:font_ptr}}
        self.aliases = {}  # aliases = {alias:font_path}
        self._textcolor = None
        self._bgcolor = None
        self.color = color
        self.bg_color = bg_color
        self.size = size
        self._default_font = self.add(font_path, alias, size, index)

    def __del__(self):
        """Close all opened fonts."""
        self.close()

    def close(self):
        """Closes all fonts opened by the class."""
        ttf_was_closed = _ttf_init_count > self._ttf_init_count
        if sdlttf.TTF_WasInit() > 0 and not ttf_was_closed:
            for alias, fonts in self.fonts.items():
                for size, font in fonts.items():
                    if font:
                        sdlttf.TTF_CloseFont(font)
        self.fonts = {}
        self.aliases = {}

    def add(self, font_path, alias=None, size=None, index=0):
        """Adds a font to the :class:`FontManager`.
        
        Args:
            font_path (str): The relative (or absolute) path to the font
                to load.
            alias (str, optional): The name to give the font within the
                FontManager. Defaults to the font filename if not specified.
            size (int, optional): The size (in pt) at which to load the font.
                Defaults to the FontManager's default size if not specified.
            index (int, optional): The index of the font face to load if the
                font file contains multiple faces. Defaults to 0 (first face in
                the file) if not specified.

        Returns:
            :obj:`~sdl2.sdlttf.TTF_Font`: A pointer to the ctypes font object
            for the added font.

        """
        size = size or self.size
        if alias is None:
            # If no alias given, take the font name as alias
            basename = os.path.basename(font_path)
            alias = os.path.splitext(basename)[0]
            if alias in self.fonts:
                if size in self.fonts[alias] and self.fonts[alias]:
                    # font with selected size already opened
                    return
                else:
                    self._change_font_size(alias, size)
                    return
            else:
                if not os.path.isfile(font_path):
                    raise IOError("Cannot find %s" % font_path)

        font = self._load_font(font_path, size, index)
        self.aliases[alias] = font_path
        self.fonts[alias] = {}
        self.fonts[alias][size] = font
        return font

    def _load_font(self, font_path, size, index=0):
        """Helper function to open the font.

        Raises an exception if something went wrong.
        """
        fullpath, fname = _validate_path(font_path, "a font")
        fullpath = byteify(fullpath)
        font = sdlttf.TTF_OpenFontIndex(fullpath, size, index)
        if not font:
            raise_sdl_err("opening the font '{0}'".format(fname))
        return font

    def _change_font_size(self, alias, size):
        """Loads an already opened font in another size."""
        if alias not in self.fonts:
            raise KeyError("Font %s not loaded in FontManager" % alias)
        font = self._load_font(self.aliases[alias], size)
        self.fonts[alias][size] = font

    @property
    def color(self):
        """:obj:`~sdl2.ext.Color`: The color to use for rendering text."""
        c = self._textcolor
        return Color(c.r, c.g, c.b, c.a)

    @color.setter
    def color(self, value):
        c = convert_to_color(value)
        self._textcolor = pixels.SDL_Color(c.r, c.g, c.b, c.a)

    @property
    def bg_color(self):
        """:obj:`~sdl2.ext.Color`: The background color to use for rendering."""
        c = self._bgcolor
        return Color(c.r, c.g, c.b, c.a)

    @bg_color.setter
    def bg_color(self, value):
        c = convert_to_color(value)
        self._bgcolor = pixels.SDL_Color(c.r, c.g, c.b, c.a)

    @property
    def default_font(self):
        """str: The name of the default font. Must be set to the alias of a
        currently-loaded font.

        """
        for alias in self.fonts:
            for size, font in self.fonts[alias].items():
                if font == self._default_font:
                    return alias

    @default_font.setter
    def default_font(self, value):
        alias = value
        size = self.size
        if alias not in self.fonts:
            raise ValueError("Font %s not loaded in FontManager" % alias)
        # Check if size is already loaded, otherwise do it.
        if size not in self.fonts[alias]:
            self._change_font_size(alias, size)
            size = list(self.fonts[alias].keys())[0]
        self._default_font = self.fonts[alias][size]

    def render(self, text, alias=None, size=None, width=None, color=None,
               bg_color=None, **kwargs):
        """Renders text to a surface.

        Args:
            text (str): The text to render.
            alias (str, optional): The alias of the font to use for rendering
                the text. Defaults to the FontManager's default font if not
                specified.
            size (int, optional): The size (in pt) at which to render the font.
                Defaults to the FontManager's default size if not specified.
            width (int, optional): The width (in pixels) of the output surface.
                If a line of text exceeds this value, it will be automatically
                wrapped to fit within the specified width. Defaults to ``None``.
            color (~sdl2.ext.Color): The font rendering color. Defaults to the
                FontManager's default color if not specified.
            bg_color (~sdl2.ext.Color, optional): The background surface color.
                Defaults to the FontManager's default background color if not
                specified.

        Returns:
            :obj:`~sdl2.SDL_Surface`: A 32-bit ARGB surface containing the
            rendered text.

        """
        alias = alias or self.default_font
        size = size or self.size
        if bg_color is None:
            bg_color = self._bgcolor
        elif not isinstance(bg_color, pixels.SDL_Color):
            c = convert_to_color(bg_color)
            bg_color = pixels.SDL_Color(c.r, c.g, c.b, c.a)
        if color is None:
            color = self._textcolor
        elif not isinstance(color, pixels.SDL_Color):
            c = convert_to_color(color)
            color = pixels.SDL_Color(c.r, c.g, c.b, c.a)
        if len(self.fonts) == 0:
            raise TypeError("There are no fonts selected.")
        font = self._default_font
        if alias not in self.aliases:
            raise KeyError("Font %s not loaded" % font)
        elif size not in self.fonts[alias]:
            self._change_font_size(alias, size)
        font = self.fonts[alias][size]
        text = byteify(text, "utf-8")
        if width:
            fontsf = sdlttf.TTF_RenderUTF8_Blended_Wrapped(font, text, color,
                                                           width)
            if not fontsf:
                raise_sdl_err("rendering the text")
            if bg_color != pixels.SDL_Color(0, 0, 0):
                fontsf = fontsf.contents
                w, h = fontsf.w, fontsf.h
                bpp = fontsf.format.contents.BitsPerPixel
                fmt = fontsf.format.contents.format
                bgsf = surface.SDL_CreateRGBSurfaceWithFormat(0, w, h, bpp, fmt)
                if not bgsf:
                    surface.SDL_FreeSurface(fontsf)
                    raise_sdl_err("creating the background surface")
                bg_color = prepare_color(bg_color, bgsf.contents)
                surface.SDL_FillRect(bgsf, None, bg_color)
                surface.SDL_BlitSurface(fontsf, None, bgsf, None)
                return bgsf.contents
            return fontsf.contents
        sf = None
        if bg_color == pixels.SDL_Color(0, 0, 0):
            sf = sdlttf.TTF_RenderUTF8_Blended(font, text, color)
        else:
            sf = sdlttf.TTF_RenderUTF8_Shaded(font, text, color,
                                              bg_color)
        if not sf:
            raise_sdl_err("rendering the text")
        return sf.contents



import sys
import pytest
import sdl2
from sdl2 import cpuinfo

_HASMP = True
try:
    import multiprocessing
except:
    _HASMP = False


def test_SDL_GetCPUCacheLineSize():
    ret = cpuinfo.SDL_GetCPUCacheLineSize()
    assert isinstance(ret, int)

def test_SDL_GetCPUCount():
    if _HASMP:
        assert multiprocessing.cpu_count() == cpuinfo.SDL_GetCPUCount()
    else:
        assert cpuinfo.SDL_GetCPUCount() >= 1

def test_SDL_Has3DNow():
    ret = cpuinfo.SDL_Has3DNow()
    assert ret in (0, 1)

def test_SDL_HasAltiVec():
    ret = cpuinfo.SDL_HasAltiVec()
    assert ret in (0, 1)

def test_SDL_HasMMX():
    ret = cpuinfo.SDL_HasMMX()
    assert ret in (0, 1)

def test_SDL_HasRDTSC():
    ret = cpuinfo.SDL_HasRDTSC()
    assert ret in (0, 1)

def test_SDL_HasSSE():
    ret = cpuinfo.SDL_HasSSE()
    assert ret in (0, 1)

def test_SDL_HasSSE2():
    ret = cpuinfo.SDL_HasSSE2()
    assert ret in (0, 1)

def test_SDL_HasSSE3():
    ret = cpuinfo.SDL_HasSSE3()
    assert ret in (0, 1)

def test_SDL_HasSSE41():
    ret = cpuinfo.SDL_HasSSE41()
    assert ret in (0, 1)

def test_SDL_HasSSE42():
    ret = cpuinfo.SDL_HasSSE42()
    assert ret in (0, 1)

def test_SDL_HasAVX():
    ret = cpuinfo.SDL_HasAVX()
    assert ret in (0, 1)

def test_SDL_HasAVX2():
    ret = cpuinfo.SDL_HasAVX2()
    assert ret in (0, 1)

def test_SDL_GetSystemRAM():
    ret = cpuinfo.SDL_GetSystemRAM()
    assert ret > 0

@pytest.mark.skipif(sdl2.dll.version < 2009, reason="not available")
def test_SDL_HasAVX512F():
    ret = cpuinfo.SDL_HasAVX512F()
    assert ret in (0, 1)

@pytest.mark.skipif(sdl2.dll.version < 2012, reason="not available")
def test_SDL_HasARMSIMD():
    ret = cpuinfo.SDL_HasARMSIMD()
    assert ret in (0, 1)

@pytest.mark.skipif(sdl2.dll.version < 2006, reason="not available")
def test_SDL_HasNEON():
    ret = cpuinfo.SDL_HasNEON()
    assert ret in (0, 1)

@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_SIMDGetAlignment():
    ret = cpuinfo.SDL_SIMDGetAlignment()
    assert ret % 8 == 0 # Should be multiple of 8

@pytest.mark.skip("not implemented (no clue how)")
@pytest.mark.skipif(sdl2.dll.version < 2010, reason="not available")
def test_SDL_SIMDAllocFree():
    # Should test both SDL_SIMDAlloc and SDL_SIMDFree
    pass

@pytest.mark.skip("not implemented (no clue how)")
@pytest.mark.skipif(sdl2.dll.version < 2014, reason="not available")
def test_SDL_SIMDRealloc():
    pass



import sys
import math
import pytest
from sdl2 import endian


def test_SDL_BYTEORDER():
    if sys.byteorder == "little":
        assert endian.SDL_BYTEORDER == endian.SDL_LIL_ENDIAN
    else:
        assert endian.SDL_BYTEORDER == endian.SDL_BIG_ENDIAN

def test_SDL_Swap16():
    assert endian.SDL_Swap16(0xFF00) == 0x00FF
    assert endian.SDL_Swap16(0x0001) == 0x0100
    assert endian.SDL_Swap16(0x0032) == 0x3200
    assert endian.SDL_Swap16(0x0FF0) == 0xF00F
    assert endian.SDL_Swap16(0x00FF) == 0xFF00
    assert endian.SDL_Swap16(0x1234) == 0x3412
    if sys.byteorder == "little":
        assert endian.SDL_Swap16 == endian.SDL_SwapBE16
        assert endian.SDL_Swap16 != endian.SDL_SwapLE16
    else:
        assert endian.SDL_Swap16 != endian.SDL_SwapBE16
        assert endian.SDL_Swap16 == endian.SDL_SwapLE16

def test_SDL_Swap32():
    assert endian.SDL_Swap32(0xFF000000) == 0x000000FF
    assert endian.SDL_Swap32(0x00FF0000) == 0x0000FF00
    assert endian.SDL_Swap32(0x0000FF00) == 0x00FF0000
    assert endian.SDL_Swap32(0x000000FF) == 0xFF000000
    assert endian.SDL_Swap32(0x12345678) == 0x78563412
    assert endian.SDL_Swap32(0xFF00FF00) == 0x00FF00FF
    if sys.byteorder == "little":
        assert endian.SDL_Swap32 == endian.SDL_SwapBE32
        assert endian.SDL_Swap32 != endian.SDL_SwapLE32
    else:
        assert endian.SDL_Swap32 != endian.SDL_SwapBE32
        assert endian.SDL_Swap32 == endian.SDL_SwapLE32

def test_SDL_Swap64():
    assert endian.SDL_Swap64(0xFF00000000000000) == 0x00000000000000FF
    assert endian.SDL_Swap64(0x00FF000000000000) == 0x000000000000FF00
    assert endian.SDL_Swap64(0x0000FF0000000000) == 0x0000000000FF0000
    assert endian.SDL_Swap64(0x000000FF00000000) == 0x00000000FF000000
    assert endian.SDL_Swap64(0x00000000FF000000) == 0x000000FF00000000
    assert endian.SDL_Swap64(0x0000000000FF0000) == 0x0000FF0000000000
    assert endian.SDL_Swap64(0x000000000000FF00) == 0x00FF000000000000
    assert endian.SDL_Swap64(0x00000000000000FF) == 0xFF00000000000000
    assert endian.SDL_Swap64(0x0123456789ABCDEF) == 0xEFCDAB8967452301
    if sys.byteorder == "little":
        assert endian.SDL_Swap64 == endian.SDL_SwapBE64
        assert endian.SDL_Swap64 != endian.SDL_SwapLE64
    else:
        assert endian.SDL_Swap64 != endian.SDL_SwapBE64
        assert endian.SDL_Swap64 == endian.SDL_SwapLE64

def test_SDL_SwapFloat():
    v = -100.0
    while v < 101:
        p = endian.SDL_SwapFloat(v)
        assert p != v
        assert endian.SDL_SwapFloat(p) == v
        v += 0.8
    values = (
        sys.float_info.epsilon,
        sys.float_info.min,
        sys.float_info.max,
        -sys.float_info.min,
        math.pi,
        -math.pi
    )
    for v in values:
        p = endian.SDL_SwapFloat(v)
        assert p != v
        assert endian.SDL_SwapFloat(p) == v
    if sys.byteorder == "little":
        assert endian.SDL_SwapFloat == endian.SDL_SwapFloatBE
        assert endian.SDL_SwapFloat != endian.SDL_SwapFloatLE
    else:
        assert endian.SDL_SwapFloat != endian.SDL_SwapFloatBE
        assert endian.SDL_SwapFloat == endian.SDL_SwapFloatLE



import sys
import pytest
import copy
from ctypes import c_int, POINTER, byref, cast, ArgumentError
import sdl2
from sdl2 import SDL_Init, SDL_Quit, SDL_INIT_EVERYTHING, SDL_TRUE
from sdl2 import pixels
from sdl2.pixels import SDL_Color
from sdl2.stdinc import Uint8, Uint16, Uint32

RGBA32 = [0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF]
RGBX32 = [0xFF000000, 0x00FF0000, 0x0000FF00, 0]
if sys.byteorder == "big":
    RGBA32.reverse()
    RGBX32.reverse()


# Test custom macros

def test_SDL_FOURCC():
    assert pixels.SDL_FOURCC("0", "0", "0", "0") == 0x30303030
    assert pixels.SDL_FOURCC("1", "1", "1", "1") == 0x31313131
    assert pixels.SDL_FOURCC("1", "2", "3", "4") == 0x34333231
    assert pixels.SDL_FOURCC("4", "3", "2", "1") == 0x31323334
    with pytest.raises(TypeError):
        pixels.SDL_FOURCC("a", "a", "a", None)
    with pytest.raises(TypeError):
        pixels.SDL_FOURCC("a", "a", "a", 1)

def test_SDL_DEFINE_PIXELFORMAT():
    fmt = pixels.SDL_DEFINE_PIXELFORMAT(
        pixels.SDL_PIXELTYPE_INDEX1, pixels.SDL_BITMAPORDER_4321, 0, 1, 0
    )
    assert fmt == pixels.SDL_PIXELFORMAT_INDEX1LSB

    fmt = pixels.SDL_DEFINE_PIXELFORMAT(
        pixels.SDL_PIXELTYPE_PACKED16,
        pixels.SDL_PACKEDORDER_XRGB,
        pixels.SDL_PACKEDLAYOUT_4444,
        12, 2
    )
    assert fmt == pixels.SDL_PIXELFORMAT_RGB444

    fmt = pixels.SDL_DEFINE_PIXELFORMAT(
        pixels.SDL_PIXELTYPE_PACKED16,
        pixels.SDL_PACKEDORDER_ABGR,
        pixels.SDL_PACKEDLAYOUT_1555,
        16, 2
    )
    assert fmt == pixels.SDL_PIXELFORMAT_ABGR1555

def test_SDL_PIXELTYPE():
    expected = [
        (pixels.SDL_PIXELFORMAT_INDEX1LSB, pixels.SDL_PIXELTYPE_INDEX1),
        (pixels.SDL_PIXELFORMAT_INDEX1MSB, pixels.SDL_PIXELTYPE_INDEX1),
        (pixels.SDL_PIXELFORMAT_INDEX4LSB, pixels.SDL_PIXELTYPE_INDEX4),
        (pixels.SDL_PIXELFORMAT_ARGB8888, pixels.SDL_PIXELTYPE_PACKED32)
    ]
    for fmt, pxtype in expected:
        assert pixels.SDL_PIXELTYPE(fmt) == pxtype

def test_SDL_PIXELORDER():
    expected = [
        (pixels.SDL_PIXELFORMAT_INDEX1LSB, pixels.SDL_BITMAPORDER_4321),
        (pixels.SDL_PIXELFORMAT_INDEX1MSB, pixels.SDL_BITMAPORDER_1234),
        (pixels.SDL_PIXELFORMAT_INDEX4LSB, pixels.SDL_BITMAPORDER_4321),
        (pixels.SDL_PIXELFORMAT_ARGB8888, pixels.SDL_PACKEDORDER_ARGB)
    ]
    for fmt, pxorder in expected:
        assert pixels.SDL_PIXELORDER(fmt) == pxorder

def test_SDL_PIXELLAYOUT():
    expected = [
        (pixels.SDL_PIXELFORMAT_INDEX1LSB, pixels.SDL_PACKEDLAYOUT_NONE),
        (pixels.SDL_PIXELFORMAT_RGB332, pixels.SDL_PACKEDLAYOUT_332),
        (pixels.SDL_PIXELFORMAT_ARGB8888, pixels.SDL_PACKEDLAYOUT_8888)
    ]
    for fmt, pxlayout in expected:
        assert pixels.SDL_PIXELLAYOUT(fmt) == pxlayout

def test_SDL_BITSPERPIXEL():
    bitspp = pixels.SDL_BITSPERPIXEL
    assert bitspp(pixels.SDL_PIXELFORMAT_INDEX1LSB) == 1
    assert bitspp(pixels.SDL_PIXELFORMAT_INDEX4LSB) == 4
    assert bitspp(pixels.SDL_PIXELFORMAT_RGB332) == 8
    assert bitspp(pixels.SDL_PIXELFORMAT_ARGB8888) == 32

def test_SDL_BYTESPERPIXEL():
    bytespp = pixels.SDL_BYTESPERPIXEL
    assert bytespp(pixels.SDL_PIXELFORMAT_INDEX1LSB) == 0
    assert bytespp(pixels.SDL_PIXELFORMAT_INDEX4LSB) == 0
    assert bytespp(pixels.SDL_PIXELFORMAT_RGB332) == 1
    assert bytespp(pixels.SDL_PIXELFORMAT_ARGB8888) == 4
    assert bytespp(pixels.SDL_PIXELFORMAT_YUY2) == 2
    assert bytespp(pixels.SDL_PIXELFORMAT_IYUV) == 1
    assert bytespp(pixels.SDL_PIXELFORMAT_UYVY) == 2

def test_SDL_ISPIXELFORMAT_INDEXED():
    isindexed = pixels.SDL_ISPIXELFORMAT_INDEXED
    assert isindexed(pixels.SDL_PIXELFORMAT_INDEX1LSB)
    assert isindexed(pixels.SDL_PIXELFORMAT_INDEX1MSB)
    assert isindexed(pixels.SDL_PIXELFORMAT_INDEX4LSB)
    assert isindexed(pixels.SDL_PIXELFORMAT_INDEX4MSB)
    assert isindexed(pixels.SDL_PIXELFORMAT_INDEX8)
    assert not isindexed(pixels.SDL_PIXELFORMAT_RGB332)
    assert not isindexed(pixels.SDL_PIXELFORMAT_ARGB8888)
    assert not isindexed(pixels.SDL_PIXELFORMAT_YUY2)

def test_SDL_ISPIXELFORMAT_ALPHA():
    isalpha = pixels.SDL_ISPIXELFORMAT_ALPHA
    assert isalpha(pixels.SDL_PIXELFORMAT_ARGB8888)
    assert isalpha(pixels.SDL_PIXELFORMAT_RGBA8888)
    assert isalpha(pixels.SDL_PIXELFORMAT_RGBA4444)
    assert isalpha(pixels.SDL_PIXELFORMAT_ABGR1555)
    assert not isalpha(pixels.SDL_PIXELFORMAT_INDEX1LSB)
    assert not isalpha(pixels.SDL_PIXELFORMAT_INDEX4MSB)
    assert not isalpha(pixels.SDL_PIXELFORMAT_RGB332)
    assert not isalpha(pixels.SDL_PIXELFORMAT_YUY2)
    assert not isalpha(pixels.SDL_PIXELFORMAT_RGBX8888)

def test_SDL_ISPIXELFORMAT_FOURCC():
    isfourcc = pixels.SDL_ISPIXELFORMAT_FOURCC
    assert isfourcc(pixels.SDL_PIXELFORMAT_YV12)
    assert isfourcc(pixels.SDL_PIXELFORMAT_IYUV)
    assert isfourcc(pixels.SDL_PIXELFORMAT_YUY2)
    assert isfourcc(pixels.SDL_PIXELFORMAT_UYVY)
    assert isfourcc(pixels.SDL_PIXELFORMAT_YVYU)
    assert not isfourcc(pixels.SDL_PIXELFORMAT_ARGB8888)
    assert not isfourcc(pixels.SDL_PIXELFORMAT_ARGB4444)
    assert not isfourcc(pixels.SDL_PIXELFORMAT_INDEX8)


# Test pixel format functions

def test_SDL_GetPixelFormatName():
    expected = [
        (pixels.SDL_PIXELFORMAT_INDEX1LSB, b"SDL_PIXELFORMAT_INDEX1LSB"),
        (pixels.SDL_PIXELFORMAT_UNKNOWN, b"SDL_PIXELFORMAT_UNKNOWN"),
        (pixels.SDL_PIXELFORMAT_UYVY, b"SDL_PIXELFORMAT_UYVY"),
        (99999, b"SDL_PIXELFORMAT_UNKNOWN")
    ]
    for fmt, name in expected:
        assert pixels.SDL_GetPixelFormatName(fmt) == name

def test_SDL_MasksToPixelFormatEnum():
    formats = [
        (pixels.SDL_PIXELFORMAT_RGBA8888, [32] + RGBA32),
        (pixels.SDL_PIXELFORMAT_RGBX8888, [32] + RGBX32),
        (pixels.SDL_PIXELFORMAT_INDEX1MSB, [1, 0, 0, 0, 0]),
    ]
    for fmt, expected in formats:
        bpp, r, g, b, a = expected
        enum = pixels.SDL_MasksToPixelFormatEnum(bpp, r, g, b, a)
        assert enum == fmt

    enum = pixels.SDL_MasksToPixelFormatEnum(17, 3, 6, 64, 255)
    assert enum == pixels.SDL_PIXELFORMAT_UNKNOWN

def test_SDL_PixelFormatEnumToMasks():
    formats = [
        (pixels.SDL_PIXELFORMAT_RGBA8888, [32] + RGBA32),
        (pixels.SDL_PIXELFORMAT_RGBX8888, [32] + RGBX32),
        (pixels.SDL_PIXELFORMAT_INDEX1LSB, [1, 0, 0, 0, 0]),
        (pixels.SDL_PIXELFORMAT_INDEX1MSB, [1, 0, 0, 0, 0]),
        (pixels.SDL_PIXELFORMAT_UNKNOWN, [0, 0, 0, 0, 0]),
    ]
    bpp = c_int(0)
    r, g, b, a = Uint32(0), Uint32(0), Uint32(0), Uint32(0)
    for fmt, expected in formats:
        ret = pixels.SDL_PixelFormatEnumToMasks(
            fmt, byref(bpp), byref(r), byref(g), byref(b), byref(a)
        )
        assert sdl2.SDL_GetError() == b""
        assert ret == SDL_TRUE
        assert [x.value for x in (bpp, r, g, b, a)] == expected

def test_SDL_AllocFreeFormat():
    formats = [
        [pixels.SDL_PIXELFORMAT_RGBA8888, (32, 4)],
        [pixels.SDL_PIXELFORMAT_INDEX1LSB, (1, 1)],
        [pixels.SDL_PIXELFORMAT_INDEX4MSB, (4, 1)],
        [pixels.SDL_PIXELFORMAT_ARGB4444, (16, 2)],
    ]
    for fmt, info in formats:
        bits_pp, bytes_pp = info
        pixfmt = pixels.SDL_AllocFormat(fmt)
        assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
        assert pixfmt.contents.format == fmt
        assert pixfmt.contents.BitsPerPixel == bits_pp
        assert pixfmt.contents.BytesPerPixel == bytes_pp
        pixels.SDL_FreeFormat(pixfmt)

def test_SDL_AllocFreePalette():
    palette = pixels.SDL_AllocPalette(10)
    assert isinstance(palette.contents, pixels.SDL_Palette)
    assert palette.contents.ncolors == 10

    colors = palette.contents.colors
    for x in range(palette.contents.ncolors):
        assert isinstance(colors[x], SDL_Color)
    colors[3].r = 70
    assert colors[3].r == 70

    # Try modifying the palette
    color = colors[4]
    color.g = 33
    assert color.g == 33
    assert colors[4].g == 33

    pixels.SDL_FreePalette(palette)

def test_SDL_SetPaletteColors():
    colors = []
    for v in range(20):
        colors.append(SDL_Color(v, v + 10, v + 20))
    col_array = (SDL_Color * len(colors))(*colors)

    palette = pixels.SDL_AllocPalette(20)
    assert isinstance(palette.contents, pixels.SDL_Palette)
    assert palette.contents.ncolors == 20
    
    pixels.SDL_SetPaletteColors(palette, col_array, 0, 20)
    for index in range(20):
        rgb = palette.contents.colors[index]
        assert rgb == colors[index]

    pixels.SDL_FreePalette(palette)

def test_SDL_SetPixelFormatPalette():
    palette = pixels.SDL_AllocPalette(2)
    assert isinstance(palette.contents, pixels.SDL_Palette)
    palette.contents.colors[1].r = 128
    palette.contents.colors[1].g = 64
    palette.contents.colors[1].b = 0

    pixfmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_INDEX1MSB)
    assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
    ret = pixels.SDL_SetPixelFormatPalette(pixfmt, palette)
    assert sdl2.SDL_GetError() == b""
    assert ret == 0

    fmt_palette = pixfmt.contents.palette.contents
    assert fmt_palette.colors[1].r == 128
    assert fmt_palette.colors[1].g == 64
    assert fmt_palette.colors[1].b == 0

    pixels.SDL_FreeFormat(pixfmt)
    pixels.SDL_FreePalette(palette)

def test_SDL_CalculateGammaRamp():
    vals = (Uint16 * 256)()
    pixels.SDL_CalculateGammaRamp(1.0, cast(vals, POINTER(Uint16)))
    assert len(vals) == 256
    p = 0
    for x in vals:
        assert x == p
        p += 257

    pixels.SDL_CalculateGammaRamp(0.0, cast(vals, POINTER(Uint16)))
    assert len(vals) == 256
    for x in vals:
        assert x == 0

    gamma = 0.5
    pixels.SDL_CalculateGammaRamp(gamma, cast(vals, POINTER(Uint16)))
    assert len(vals) == 256
    for i in range(len(vals)):
        expected = int((i / 256.0) ** gamma * 65535 + 0.5)
        vals[i] == expected

    # Test error on negative value
    sdl2.SDL_ClearError()
    pixels.SDL_CalculateGammaRamp(-1.0, cast(vals, POINTER(Uint16)))
    assert sdl2.SDL_GetError() != b""
    sdl2.SDL_ClearError()

def test_SDL_GetRGB():
    colors = [
        (0xFFAA8811, [0xFF, 0xAA, 0x88]),
        (0x00000000, [0x00, 0x00, 0x00]),
        (0xFFFFFFFF, [0xFF, 0xFF, 0xFF]),
        (0x11223344, [0x11, 0x22, 0x33]),
    ]
    pixfmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_RGBA8888)
    assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
    r, g, b = Uint8(0), Uint8(0), Uint8(0)
    for col, expected in colors:
        pixels.SDL_GetRGB(col, pixfmt, byref(r), byref(g), byref(b))
        assert [x.value for x in (r, g, b)] == expected
    pixels.SDL_FreeFormat(pixfmt)

    # With a pallete format, first argument is index of color to retrieve
    palette = pixels.SDL_AllocPalette(2)
    palette.contents.colors[1].r = 128
    palette.contents.colors[1].g = 64
    palette.contents.colors[1].b = 32
    pixfmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_INDEX1MSB)
    assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
    pixels.SDL_SetPixelFormatPalette(pixfmt, palette)
    pixels.SDL_GetRGB(1, pixfmt, byref(r), byref(g), byref(b))
    assert [x.value for x in (r, g, b)] == [128, 64, 32]

    # With a pallete format, missing index returns all zeros
    pixels.SDL_GetRGB(16, pixfmt, byref(r), byref(g), byref(b))
    assert [x.value for x in (r, g, b)] == [0, 0, 0]
    pixels.SDL_FreeFormat(pixfmt)
    pixels.SDL_FreePalette(palette)

def test_SDL_GetRGBA():
    colors = [
        (0xFFAA8811, [0xFF, 0xAA, 0x88, 0x11]),
        (0x00000000, [0x00, 0x00, 0x00, 0x00]),
        (0xFFFFFFFF, [0xFF, 0xFF, 0xFF, 0xFF]),
        (0x11223344, [0x11, 0x22, 0x33, 0x44]),
    ]
    pixfmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_RGBA8888)
    assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
    r, g, b, a = Uint8(0), Uint8(0), Uint8(0), Uint8(0)
    for col, expected in colors:
        pixels.SDL_GetRGBA(col, pixfmt, byref(r), byref(g), byref(b), byref(a))
        assert [x.value for x in (r, g, b, a)] == expected
    pixels.SDL_FreeFormat(pixfmt)

    # With a pallete format, first argument is index of color to retrieve
    palette = pixels.SDL_AllocPalette(2)
    palette.contents.colors[1].r = 128
    palette.contents.colors[1].g = 64
    palette.contents.colors[1].b = 32
    palette.contents.colors[1].a = 96
    pixfmt = pixels.SDL_AllocFormat(pixels.SDL_PIXELFORMAT_INDEX1MSB)
    assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
    pixels.SDL_SetPixelFormatPalette(pixfmt, palette)
    pixels.SDL_GetRGBA(1, pixfmt, byref(r), byref(g), byref(b), byref(a))
    assert [x.value for x in (r, g, b, a)] == [128, 64, 32, 96]

    # With a pallete format, missing index returns all zeros
    pixels.SDL_GetRGBA(16, pixfmt, byref(r), byref(g), byref(b), byref(a))
    assert [x.value for x in (r, g, b, a)] == [0, 0, 0, 0]
    pixels.SDL_FreeFormat(pixfmt)

def test_SDL_MapRGB():
    r, g, b = (0xFF, 0xAA, 0x88)
    formats = [
        (pixels.SDL_PIXELFORMAT_RGBA8888, 0xFFAA88FF),
        (pixels.SDL_PIXELFORMAT_BGRA8888, 0x88AAFFFF),
        (pixels.SDL_PIXELFORMAT_ABGR8888, 0xFF88AAFF),
    ]
    for fmt, expected in formats:
        pixfmt = pixels.SDL_AllocFormat(fmt)
        assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
        col = pixels.SDL_MapRGB(pixfmt, r, g, b)
        assert col == expected
        pixels.SDL_FreeFormat(pixfmt)

def test_SDL_MapRGBA():
    r, g, b, a = (0xFF, 0xAA, 0x88, 0x11)
    formats = [
        (pixels.SDL_PIXELFORMAT_RGBA8888, 0xFFAA8811),
        (pixels.SDL_PIXELFORMAT_BGRA8888, 0x88AAFF11),
        (pixels.SDL_PIXELFORMAT_ABGR8888, 0x1188AAFF),
    ]
    for fmt, expected in formats:
        pixfmt = pixels.SDL_AllocFormat(fmt)
        assert isinstance(pixfmt.contents, pixels.SDL_PixelFormat)
        col = pixels.SDL_MapRGBA(pixfmt, r, g, b, a)
        assert col == expected
        pixels.SDL_FreeFormat(pixfmt)


# Test structs and objects

def test_SDL_PixelFormat():
    pixfmt = pixels.SDL_PixelFormat()
    assert isinstance(pixfmt, pixels.SDL_PixelFormat)

def test_SDL_Palette():
    palette = pixels.SDL_Palette()
    assert isinstance(palette, pixels.SDL_Palette)


class TestSDLColor(object):
    __tags__ = ["sdl"]

    def test_init(self):
        col = SDL_Color()

        # Check default values
        assert col.r == 255
        assert col.g == 255
        assert col.b == 255
        assert col.a == 255

        # Test RGB and RGBA inputs
        col_rgb = SDL_Color(128, 128, 128)
        col_rgba = SDL_Color(128, 128, 128, 64)

        # Test errors on bad values
        with pytest.raises(TypeError):
            SDL_Color(0.1, 0, 0)

    def test___repr__(self):
        c1 = SDL_Color()
        assert "SDL_Color(r=255, g=255, b=255, a=255)" == repr(c1)
        c2 = eval(repr(c1))
        assert c2 == c1
        c3 = eval(repr(c2))
        assert c3 == c2

    def test___copy__(self):
        c = SDL_Color(10, 20, 30)
        c2 = copy.copy(c)
        assert c == c2
        c.r = 100
        assert c != c2

    def test___eq__(self):
        assert SDL_Color(255, 0, 0, 0) == SDL_Color(255, 0, 0, 0)
        assert SDL_Color(0, 255, 0, 0) == SDL_Color(0, 255, 0, 0)
        assert SDL_Color(0, 0, 255, 0) == SDL_Color(0, 0, 255, 0)
        assert SDL_Color(0, 0, 0, 255) == SDL_Color(0, 0, 0, 255)
        assert SDL_Color(0, 0, 0, 0) == SDL_Color(0, 0, 0, 0)

        assert not (SDL_Color(0, 0, 0, 0) == SDL_Color(255, 0, 0, 0))
        assert not (SDL_Color(0, 0, 0, 0) == SDL_Color(0, 255, 0, 0))
        assert not (SDL_Color(0, 0, 0, 0) == SDL_Color(0, 0, 255, 0))
        assert not (SDL_Color(0, 0, 0, 0) == SDL_Color(0, 0, 0, 255))

    def test___ne__(self):
        assert SDL_Color(0, 0, 0, 0) != SDL_Color(255, 0, 0, 0)
        assert SDL_Color(0, 0, 0, 0) != SDL_Color(0, 255, 0, 0)
        assert SDL_Color(0, 0, 0, 0) != SDL_Color(0, 0, 255, 0)
        assert SDL_Color(0, 0, 0, 0) != SDL_Color(0, 0, 255, 0)
        assert SDL_Color(0, 0, 0, 0) != SDL_Color(0, 0, 0, 255)

        assert not (SDL_Color(255, 0, 0, 0) != SDL_Color(255, 0, 0, 0))
        assert not (SDL_Color(0, 255, 0, 0) != SDL_Color(0, 255, 0, 0))
        assert not (SDL_Color(0, 0, 255, 0) != SDL_Color(0, 0, 255, 0))
        assert not (SDL_Color(0, 0, 0, 255) != SDL_Color(0, 0, 0, 255))

    def test_colors(self):
        col = SDL_Color()

        # Check getting/setting values
        for x in (0, 16, 32, 64, 128, 255):
            col.r = x
            col.g = x
            col.b = x
            col.a = x
            assert col.r == x
            assert col.g == x
            assert col.b == x
            assert col.a == x

        # Test wrap for out-of-bounds values
        col.r = 256
        assert col.r == 0
        col.g = -1
        assert col.g == 255



"""Window routines to manage on-screen windows."""
from ctypes import c_int, byref
from .compat import stringify, utf8
from .common import SDLError
from .. import video

__all__ = ["Window"]


class Window(object):
    """Creates a visible window with an optional border and title text.

    In SDL2, a window is the object through which visuals are displayed and
    all input events are captured. This class is a Pythonic alternative to
    working directly with SDL2's :obj:`~sdl2.SDL_Window` type, wrapping a
    number of useful functions for working with windows.

    The created Window is hidden by default, but can be shown using the
    :meth:`~Window.show` method. Additionally, the type and properties of the
    created window can be configured using various flags:

    ================================== =========================================
    Flag                               Description
    ================================== =========================================
    ``SDL_WINDOW_SHOWN``               Will be shown when created
    ``SDL_WINDOW_HIDDEN``              Will be hidden when created
    ``SDL_WINDOW_BORDERLESS``          Will not be decorated by the OS
    ``SDL_WINDOW_RESIZABLE``           Will be resizable
    ``SDL_WINDOW_MINIMIZED``           Will be created in a minimized state
    ``SDL_WINDOW_MAXIMIZED``           Will be created in a maximized state
    ``SDL_WINDOW_FULLSCREEN``          Will be fullscreen
    ``SDL_WINDOW_FULLSCREEN_DESKTOP``  Will be fullscreen at the current desktop
                                       resolution
    ``SDL_WINDOW_OPENGL``              Will be usable with an OpenGL context
    ``SDL_WINDOW_VULKAN``              Will be usable with a Vulkan instance
    ``SDL_WINDOW_METAL``               Will be usable with a Metal context
    ``SDL_WINDOW_ALLOW_HIGHDPI``       Will be created in high-DPI mode
                                       (if supported)
    ``SDL_WINDOW_INPUT_FOCUS``         Will have input focus when created
    ``SDL_WINDOW_MOUSE_FOCUS``         Will have mouse focus when created
    ``SDL_WINDOW_INPUT_GRABBED``       Will prevent the mouse from leaving the
                                       bounds of the window
    ``SDL_WINDOW_MOUSE_CAPTURE``       Will capture mouse to track input outside
                                       of the window when created
    ================================== =========================================

    There are also a few additional window creation flags that only affect the
    X11 window manager (i.e. most distributions of Linux and BSD):

    ================================== =====================================
    Flag                               Description
    ================================== =====================================
    ``SDL_WINDOW_ALWAYS_ON_TOP``       Should stay on top of other windows
    ``SDL_WINDOW_SKIP_TASKBAR``        Should not be added to the taskbar
    ``SDL_WINDOW_UTILITY``             Should be treated as a utility window
    ``SDL_WINDOW_TOOLTIP``             Should be treated as a tooltip
    ``SDL_WINDOW_POPUP_MENU``          Should be treated as a popup menu
    ================================== =====================================

    To combine multiple flags, you can use a bitwise OR to combine two or more
    together before passing them to the `flags` argument. For example, to create
    a fullscreen window that supports an OpenGL context and is shown on
    creation, you would use::

      win_flags = (
          sdl2.SDL_WINDOW_FULLSCREEN | sdl2.SDL_WINDOW_OPENGL |
          sdl2.SDL_WINDOW_SHOWN
      )
      sdl2.ext.Window("PySDL2", (800, 600), flags=win_flags)

    Args:
        title (str): The title to use for the window.
        size (tuple): The initial size (in pixels) of the window, in the format
            `(width, height)`.
        position (tuple, optional): The initial `(x, y)` coordinates of the
            top-left corner of the window. If not specified, defaults to letting
            the system's window manager choose a location for the window.
        flags (int, optional): The window attribute flags with which the window
            will be created. Defaults to ``SDL_WINDOW_HIDDEN``.

    Attributes:
        window (:obj:`~sdl2.SDL_Window`, None): The underlying SDL2 Window
            object. If the window has been closed, this value will be `None`.

    """
    DEFAULTFLAGS = video.SDL_WINDOW_HIDDEN
    DEFAULTPOS = (video.SDL_WINDOWPOS_UNDEFINED,
                  video.SDL_WINDOWPOS_UNDEFINED)

    def __init__(self, title, size, position=None, flags=None):
        if position is None:
            position = self.DEFAULTPOS
        if flags is None:
            flags = self.DEFAULTFLAGS

        self._title = title
        self._position = position
        self._flags = flags
        self._size = size

        self.window = None
        self.create()

    def __del__(self):
        """Releases the resources of the Window, implicitly destroying the
        underlying SDL2 window."""
        self.close()

    @property
    def position(self):
        """tuple: The (x, y) pixel coordinates of the top-left corner of the
        window.

        """
        x, y = c_int(), c_int()
        video.SDL_GetWindowPosition(self.window, byref(x), byref(y))
        return x.value, y.value

    @position.setter
    def position(self, value):
        if not self.window:
            raise SDLError("The window is not available")
        video.SDL_SetWindowPosition(self.window, value[0], value[1])
        self._position = value[0], value[1]

    @property
    def title(self):
        """str: The title of the window."""
        return stringify(video.SDL_GetWindowTitle(self.window), "utf-8")

    @title.setter
    def title(self, value):
        title_bytes = utf8(value).encode('utf-8')
        video.SDL_SetWindowTitle(self.window, title_bytes)
        self._title = value

    @property
    def size(self):
        """tuple: The dimensions of the window (in pixels) in the form
        (width, height).
        
        """
        w, h = c_int(), c_int()
        video.SDL_GetWindowSize(self.window, byref(w), byref(h))
        return w.value, h.value

    @size.setter
    def size(self, value):
        video.SDL_SetWindowSize(self.window, value[0], value[1])
        self._size = value[0], value[1]

    def create(self):
        """Creates the window if it does not already exist."""
        if self.window != None:
            return
        window = video.SDL_CreateWindow(utf8(self._title).encode('utf-8'),
                                        self._position[0], self._position[1],
                                        self._size[0], self._size[1],
                                        self._flags)
        if not window:
            raise SDLError()
        self.window = window.contents

    def open(self):
        """Creates and shows the window."""
        if self.window is None:
            self.create()
        self.show()

    def close(self):
        """Closes and destroys the window.
        
        Once this method has been called, the window cannot be re-opened.
        However, you can create a new window with the same settings using the
        :meth:`create` method.

        """
        if hasattr(self, "window") and self.window != None:
            video.SDL_DestroyWindow(self.window)
            self.window = None

    def show(self):
        """Shows the window on the display.
        
        If the window is already visible, this method does nothing.

        """
        video.SDL_ShowWindow(self.window)

    def hide(self):
        """Hides the window.

        If the window is already hidden, this method does nothing.
        
        """
        video.SDL_HideWindow(self.window)

    def maximize(self):
        """Maximizes the window."""
        video.SDL_MaximizeWindow(self.window)

    def minimize(self):
        """Minimizes the window into the system's dock or task bar."""
        video.SDL_MinimizeWindow(self.window)

    def restore(self):
        """Restores a minimized or maximized window to its original state.
        
        If the window has not been minimized or maximized, this method does
        nothing.

        """
        video.SDL_RestoreWindow(self.window)

    def refresh(self):
        """Updates the window to reflect any changes made to its surface.

        .. note::
           This only needs to be called if the window surface was acquired and
           modified using :meth:`get_surface`.

        """
        video.SDL_UpdateWindowSurface(self.window)

    def get_surface(self):
        """Gets the :obj:`~sdl2.SDL_Surface` used by the window.

        This method obtains the SDL surface used by the window, which can be
        then be modified to change the contents of the window (e.g draw shapes
        or images) using functions such as :func:`sdl2.SDL_BlitSurface` or
        :func:`sdl2.ext.fill`.

        The obtained surface will be invalidated if the window is resized. As
        such, you will need to call this method again whenever this happens
        in order to continue to access the window's contents.

        .. note::
           If using OpenGL/Vulkan/Metal rendering or the SDL rendering API (e.g.
           :func:`sdl2.SDL_CreateRenderer`) for drawing to the window, this
           method should not be used.

        Returns:
            :obj:`~sdl2.SDL_Surface`: The surface associated with the window.

        """
        sf = video.SDL_GetWindowSurface(self.window)
        if not sf:
            raise SDLError()
        return sf.contents



import gc
import sys
import pytest
from ctypes import ArgumentError
from sdl2 import ext as sdl2ext
from sdl2.ext.resources import Resources
from sdl2.render import (
    SDL_TEXTUREACCESS_STATIC, SDL_TEXTUREACCESS_STREAMING, SDL_TEXTUREACCESS_TARGET
)
from sdl2.surface import SDL_Surface, SDL_CreateRGBSurface, SDL_FreeSurface
from sdl2.pixels import SDL_MapRGBA
from sdl2.error import SDL_GetError, SDL_ClearError

RESOURCES = Resources(__file__, "resources")

BLACK = (0, 0, 0, 255)
RED = (255, 0, 0, 255)
BLUE = (0, 0, 255, 255)

sprite_test_sizes = [
    (1, 1), (100, 100), (16, 32), (55, 77), (20, 4)
]


def check_pixels(surf, w, h, sprite, c1, c2, cx=0, cy=0):
    cx = cx + sprite.x
    cy = cy + sprite.y
    cw, ch = sprite.size
    cmy = cy + ch
    cmx = cx + cw
    c1 = SDL_MapRGBA(surf.format, c1[0], c1[1], c1[2], c1[3])
    c2 = [SDL_MapRGBA(surf.format, c[0], c[1], c[2], c[3]) for c in c2]
    view = sdl2ext.PixelView(surf)
    for y in range(w):
        for x in range(h):
            if cy <= y < cmy and cx <= x < cmx:
                assert view[y][x] == c1
            else:
                assert view[y][x] in c2
    del view


class TestSpriteFactory(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        factory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)
        assert isinstance(factory, sdl2ext.SpriteFactory)
        assert factory.default_args == {}

        factory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE, bananas="tasty")
        assert isinstance(factory, sdl2ext.SpriteFactory)
        assert factory.default_args == {"bananas": "tasty"}

        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)

        factory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        assert isinstance(factory, sdl2ext.SpriteFactory)

        factory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        assert isinstance(factory, sdl2ext.SpriteFactory)
        assert factory.default_args == {"renderer": renderer}

        with pytest.raises(ValueError):
            sdl2ext.SpriteFactory("Test")
        with pytest.raises(ValueError):
            sdl2ext.SpriteFactory(-456)

    def test_create_sprite(self, with_sdl):
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)
        tfactory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        sfactory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)

        for w, h in sprite_test_sizes:
            for bpp in (1, 4, 8, 12, 15, 16, 24, 32):
                sprite = sfactory.create_sprite(size=(w, h), bpp=bpp)
                assert isinstance(sprite, sdl2ext.SoftwareSprite)
                assert SDL_GetError() == b""
            sprite = tfactory.create_sprite(size=(w, h))
            assert isinstance(sprite, sdl2ext.TextureSprite)
            assert SDL_GetError() == b""

        with pytest.raises(sdl2ext.SDLError):
            tfactory.create_sprite(size=(0, 1))

    def test_create_software_sprite(self, with_sdl):
        factory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)
        for w, h in sprite_test_sizes:
            for bpp in (1, 4, 8, 12, 15, 16, 24, 32):
                sprite = factory.create_software_sprite((w, h), bpp)
                assert isinstance(sprite, sdl2ext.SoftwareSprite)
                assert SDL_GetError() == b""

        with pytest.raises(TypeError):
            factory.create_software_sprite(size=None)
        with pytest.raises(sdl2ext.SDLError):
            factory.create_software_sprite(size=(10, 10), bpp=-1)
        with pytest.raises(TypeError):
            factory.create_software_sprite(masks=5)
        with pytest.raises((ArgumentError, TypeError)):
            factory.create_software_sprite(
                size=(10, 10), masks=(None, None, None, None)
            )
        with pytest.raises((ArgumentError, TypeError)):
            factory.create_software_sprite(
                size=(10, 10), masks=("Test", 1, 2, 3)
            )

    def test_create_texture_sprite(self, with_sdl):
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)
        factory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        for w, h in sprite_test_sizes:
            sprite = factory.create_texture_sprite(renderer, size=(w, h))
            assert isinstance(sprite, sdl2ext.TextureSprite)
            assert SDL_GetError() == b""
            del sprite

        # Test different access flags
        for flag in (
            SDL_TEXTUREACCESS_STATIC,
            SDL_TEXTUREACCESS_STREAMING,
            SDL_TEXTUREACCESS_TARGET
        ):
            sprite = factory.create_texture_sprite(
                renderer, size=(64, 64), access=flag
            )
            assert isinstance(sprite, sdl2ext.TextureSprite)
            assert SDL_GetError() == b""
            del sprite

    def test_from_image(self, with_sdl):
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)
        tfactory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        sfactory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)

        for suffix in ("tif", "png", "jpg"):
            imgname = RESOURCES.get_path("surfacetest.%s" % suffix)
            tsprite = tfactory.from_image(imgname)
            assert isinstance(tsprite, sdl2ext.TextureSprite)
            ssprite = sfactory.from_image(imgname)
            assert isinstance(ssprite, sdl2ext.SoftwareSprite)

        for factory in (tfactory, sfactory):
            with pytest.raises(ValueError):
                factory.from_image(None)
            with pytest.raises(ValueError):
                factory.from_image(12345)

    @pytest.mark.skip("not implemented")
    def test_from_object(self, with_sdl):
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)
        tfactory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        sfactory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)

    def test_from_surface(self, with_sdl):
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)
        tfactory = sdl2ext.SpriteFactory(sdl2ext.TEXTURE, renderer=renderer)
        sfactory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)

        sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
        tsprite = tfactory.from_surface(sf.contents)
        assert isinstance(tsprite, sdl2ext.TextureSprite)
        ssprite = sfactory.from_surface(sf.contents)
        assert isinstance(ssprite, sdl2ext.SoftwareSprite)
        SDL_FreeSurface(sf)

        for factory in (tfactory, sfactory):
            with pytest.raises((sdl2ext.SDLError, AttributeError, ArgumentError,
                               TypeError)):
                factory.from_surface(None)
            with pytest.raises((AttributeError, ArgumentError, TypeError)):
                factory.from_surface("test")

    def test_from_text(self, with_sdl):
        sfactory = sdl2ext.SpriteFactory(sdl2ext.SOFTWARE)
        fm = sdl2ext.FontManager(RESOURCES.get_path("tuffy.ttf"))

        # No Fontmanager passed
        with pytest.raises(KeyError):
            sfactory.from_text("Test")

        # Passing various keywords arguments
        sprite = sfactory.from_text("Test", fontmanager=fm)
        assert isinstance(sprite, sdl2ext.SoftwareSprite)

        sprite = sfactory.from_text("Test", fontmanager=fm, alias="tuffy")
        assert isinstance(sprite, sdl2ext.SoftwareSprite)

        # Get text from a texture sprite factory
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.Renderer(window)
        tfactory = sdl2ext.SpriteFactory(
            sdl2ext.TEXTURE, renderer=renderer, fontmanager=fm
        )
        sprite = tfactory.from_text("Test", alias="tuffy")
        assert isinstance(sprite, sdl2ext.TextureSprite)


class TestSpriteRenderSystem(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        renderer = sdl2ext.SpriteRenderSystem()
        assert isinstance(renderer, sdl2ext.SpriteRenderSystem)
        assert renderer.sortfunc is not None
        assert sdl2ext.Sprite in renderer.componenttypes

    def test_sortfunc(self, with_sdl):
        def func(p):
            pass

        renderer = sdl2ext.SpriteRenderSystem()
        assert renderer.sortfunc is not None
        renderer.sortfunc = func
        assert renderer.sortfunc == func

        def setf(x, f):
            x.sortfunc = f
        with pytest.raises(TypeError):
            setf(renderer, None)
        with pytest.raises(TypeError):
            setf(renderer, "Test")
        with pytest.raises(TypeError):
            setf(renderer, 1234)

    @pytest.mark.skip("not implemented")
    def test_render(self, with_sdl):
        pass

    @pytest.mark.skip("not implemented")
    def test_process(self, with_sdl):
        pass


class TestSoftwareSpriteRenderSystem(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        window = sdl2ext.Window("Test", size=(10, 10))
        renderer = sdl2ext.SoftwareSpriteRenderSystem(window)
        assert isinstance(renderer, sdl2ext.SpriteRenderSystem)
        assert renderer.window == window.window
        assert isinstance(renderer.surface, SDL_Surface)

        renderer = sdl2ext.SoftwareSpriteRenderSystem(window.window)
        assert isinstance(renderer, sdl2ext.SpriteRenderSystem)
        assert renderer.window == window.window
        assert isinstance(renderer.surface, SDL_Surface)

        assert renderer.sortfunc is not None
        assert not (sdl2ext.Sprite in renderer.componenttypes)
        assert sdl2ext.SoftwareSprite in renderer.componenttypes

        with pytest.raises(TypeError):
            sdl2ext.SoftwareSpriteRenderSystem()
        with pytest.raises(TypeError):
            sdl2ext.SoftwareSpriteRenderSystem(None)

    def test_render(self, with_sdl):
        # Create two software sprites for testing
        sf1 = SDL_CreateRGBSurface(0, 12, 7, 32, 0, 0, 0, 0)
        sp1 = sdl2ext.SoftwareSprite(sf1.contents, True)
        sdl2ext.fill(sp1, RED)
        sf2 = SDL_CreateRGBSurface(0, 3, 9, 32, 0, 0, 0, 0)
        sp2 = sdl2ext.SoftwareSprite(sf2.contents, True)
        sdl2ext.fill(sp2, BLUE)
        sprites = [sp1, sp2]

        # Create a window and renderer for the tests
        window = sdl2ext.Window("Test", size=(20, 20))
        renderer = sdl2ext.SoftwareSpriteRenderSystem(window)
        assert isinstance(renderer, sdl2ext.SpriteRenderSystem)
        sdl2ext.fill(renderer.surface, BLACK)

        # Test rendering a single sprite to different locations
        surf = renderer.surface
        for x, y in ((0, 0), (3, 3), (20, 20), (1, 12), (5, 6)):
            sp1.position = x, y
            renderer.render(sp1)
            check_pixels(surf, 20, 20, sp1, RED, [BLACK])
            sdl2ext.fill(surf, BLACK)
        
        # Test rendering multiple sprites in different positions
        sp1.position = 0, 0
        sp2.position = 14, 1
        renderer.render(sprites)
        check_pixels(surf, 20, 20, sp1, RED, [BLACK, BLUE])
        check_pixels(surf, 20, 20, sp2, BLUE, [BLACK, RED])
        sdl2ext.fill(surf, BLACK)

        # Test rendering multiple sprites with an x/y offset
        renderer.render(sprites, 1, 2)
        check_pixels(surf, 20, 20, sp1, RED, [BLACK, BLUE], 1, 2)
        check_pixels(surf, 20, 20, sp2, BLUE, [BLACK, RED], 1, 2)

    def test_process(self, with_sdl):
        # Create two software sprites for testing & give them depths
        sf1 = SDL_CreateRGBSurface(0, 5, 10, 32, 0, 0, 0, 0)
        sp1 = sdl2ext.SoftwareSprite(sf1.contents, True)
        sdl2ext.fill(sp1, RED)
        sf2 = SDL_CreateRGBSurface(0, 5, 10, 32, 0, 0, 0, 0)
        sp2 = sdl2ext.SoftwareSprite(sf2.contents, True)
        sdl2ext.fill(sp2, BLUE)
        sp1.depth = 0
        sp2.depth = 99
        sprites = [sp1, sp2]

        # Create a window and renderer for the tests
        window = sdl2ext.Window("Test", size=(20, 20))
        renderer = sdl2ext.SoftwareSpriteRenderSystem(window)
        assert isinstance(renderer, sdl2ext.SpriteRenderSystem)
        sdl2ext.fill(renderer.surface, BLACK)

        # Make sure only sp2 visible on surface, since its depth is higher
        renderer.process("fakeworld", sprites)
        check_pixels(renderer.surface, 20, 20, sp1, BLUE, [BLACK])
        check_pixels(renderer.surface, 20, 20, sp2, BLUE, [BLACK])


class TestTextureSpriteRenderSystem(object):
    __tags__ = ["sdl", "sdl2ext"]

    @pytest.mark.skip("not implemented")
    def test_init(self, with_sdl):
        pass

    @pytest.mark.skip("not implemented")
    def test_render(self, with_sdl):
        pass

    @pytest.mark.skip("not implemented")
    def test_process(self, with_sdl):
        pass



import sys
import pytest
import sdl2
from sdl2 import SDL_WINDOW_SHOWN
from sdl2 import ext as sdl2ext
from sdl2 import surface, video
from .conftest import SKIP_ANNOYING

# Some tests don't work properly with some video drivers, so check the name
DRIVER_DUMMY = False
DRIVER_X11 = False
try:
    sdl2.SDL_Init(sdl2.SDL_INIT_VIDEO)
    driver_name = video.SDL_GetCurrentVideoDriver()
    sdl2.SDL_Quit()
    DRIVER_DUMMY = driver_name == b"dummy"
    DRIVER_X11 = driver_name == b"x11"
except:
    pass


class TestExtWindow(object):
    __tags__ = ["sdl", "sdl2ext"]

    def test_init(self, with_sdl):
        flags = video.SDL_WINDOW_BORDERLESS
        sizes = ((1, 1), (10, 10), (10, 20), (200, 17), (640, 480))
        for w, h in sizes:
            window = sdl2ext.Window("Window", size=(w, h), flags=flags)
            assert window.size == (w, h)
            window.close()

    def test_title(self, with_sdl):
        window = sdl2ext.Window("Window", size=(10, 10))
        assert window.title == "Window"
        window.title = b"Test1234"
        assert window.title == "Test1234"
        window.close()

    def test_show_hide(self, with_sdl):
        get_flags = video.SDL_GetWindowFlags
        window = sdl2ext.Window("Test Show Window", size=(200, 200))
        assert get_flags(window.window) & SDL_WINDOW_SHOWN != SDL_WINDOW_SHOWN
        window.show()
        assert get_flags(window.window) & SDL_WINDOW_SHOWN == SDL_WINDOW_SHOWN
        window.hide()
        assert get_flags(window.window) & SDL_WINDOW_SHOWN != SDL_WINDOW_SHOWN
        window.close()

    @pytest.mark.skipif(SKIP_ANNOYING, reason="Skip unless requested")
    def test_maximize(self, with_sdl):
        get_flags = video.SDL_GetWindowFlags
        max_flag = video.SDL_WINDOW_MAXIMIZED
        flags = video.SDL_WINDOW_RESIZABLE
        window = sdl2ext.Window("Test", size=(200, 200), flags=flags)
        window.show()
        assert get_flags(window.window) & max_flag != max_flag
        window.maximize()
        if not DRIVER_DUMMY:
            assert get_flags(window.window) & max_flag == max_flag
        window.close()

    @pytest.mark.skipif(SKIP_ANNOYING, reason="Skip unless requested")
    def test_minimize_restore(self, with_sdl):
        get_flags = video.SDL_GetWindowFlags
        min_flag = video.SDL_WINDOW_MINIMIZED
        window = sdl2ext.Window("Test", size=(200, 200))
        window.show()
        assert get_flags(window.window) & min_flag != min_flag
        window.minimize()
        if not (DRIVER_DUMMY or DRIVER_X11):
            assert get_flags(window.window) & min_flag == min_flag
        window.restore()
        assert get_flags(window.window) & min_flag != min_flag
        window.close()

    @pytest.mark.skip("not implemented")
    def test_refresh(self, with_sdl):
        pass

    def test_get_surface(self, with_sdl):
        window = sdl2ext.Window("Surface", size=(200, 200))
        sf = window.get_surface()
        assert isinstance(sf, surface.SDL_Surface)
        window.close()

    def test_open_close(self, with_sdl):
        get_flags = video.SDL_GetWindowFlags
        window = sdl2ext.Window("Test", size=(200, 200))
        window.close()
        assert window.window == None
        window.open()
        assert isinstance(window.window, video.SDL_Window)
        assert get_flags(window.window) & SDL_WINDOW_SHOWN == SDL_WINDOW_SHOWN
        window.close()

    def test_position(self, with_sdl):
        window = sdl2ext.Window("Position", size=(200, 200), position=(100, 100))
        assert window.position == (100, 100)
        window.position = 70, 300
        assert window.position == (70, 300)
        window.close()

    def test_size(self, with_sdl):
        # This may fail for fullscreen WMs or Win10 tablet modes
        flags = video.SDL_WINDOW_RESIZABLE
        window = sdl2ext.Window("Size", size=(200, 200), flags=flags)
        assert window.size == (200, 200)
        window.size = 150, 77
        assert window.size == (150, 77)
        window.close()



import os
import sys
import pytest
import sdl2
from sdl2 import ext as sdl2ext
from sdl2.ext import color
from sdl2 import surface as surf
from sdl2 import pixels

try:
    from sdl2 import sdlimage
    _HASSDLIMAGE=True
except:
    _HASSDLIMAGE=False

try:
    import PIL
    _HASPIL = True
except ImportError:
    _HASPIL = False


parent_path = os.path.abspath(os.path.dirname(__file__))
resource_path = os.path.join(parent_path, "resources")

is32bit = sys.maxsize <= 2**32
ismacos = sys.platform == "darwin"
skip_formats = []

if _HASSDLIMAGE:
    # SVG unsupported on SDL2_image < 2.0.2
    if sdlimage.dll.version < 2002:
        skip_formats.append("svg")

    # As of SDL2_image 2.0.5, XCF support seems to be broken (fails to load
    # on 32-bit, transparent surface on 64-bit)
    # XCF support is also broken in official SDL2_image macOS .frameworks
    if sdlimage.dll.version == 2005 or ismacos:
        skip_formats.append("xcf")

    # WEBP support is broken in the 32-bit Windows SDL2_image 2.0.2 binary
    if is32bit and sdlimage.dll.version == 2002:
        skip_formats.append("webp")


# List of lossy/non-color formats that shouldn't be compared against reference
# during tests
skip_color_check = ['gif', 'jpg', 'lbm', 'pbm', 'pgm', 'svg', 'webp']

# SDL 2.0.10 has a bug that messes up converting surfaces with transparency
if sdl2.dll.version == 2010:
    skip_color_check.append('xpm')

colors = {
    'red': color.Color(255, 0, 0, 255),
    'blue': color.Color(0, 0, 255, 255),
    'black': color.Color(0, 0, 0, 255),
    'white': color.Color(255, 255, 255, 255)
}

def check_image_contents(img):
    # Test different coordinates on surface
    pxview = sdl2ext.PixelView(img)
    img_red = color.ARGB(pxview[0][0])
    img_blue = color.ARGB(pxview[0][16])
    img_white = color.ARGB(pxview[0][31])
    img_black = color.ARGB(pxview[31][31])
    assert img_red == colors['red']
    assert img_blue == colors['blue']
    assert img_white == colors['white']
    assert img_black == colors['black']



def test_load_bmp(with_sdl):
    # Test loading a basic BMP image
    img_path = os.path.join(resource_path, "surfacetest.bmp")
    sf = sdl2ext.load_bmp(img_path)
    assert isinstance(sf, surf.SDL_Surface)
    check_image_contents(sf)
    surf.SDL_FreeSurface(sf)

    # Test exception on missing file
    bad_path = os.path.join(resource_path, "doesnt_exist.bmp")
    with pytest.raises(IOError):
        sdl2ext.load_bmp(bad_path)

    # Test exception on bad file type
    bad_type = os.path.join(resource_path, "surfacetest.png")
    with pytest.raises(sdl2ext.SDLError):
        sdl2ext.load_bmp(bad_type)


def test_save_bmp(with_sdl, tmpdir):
    # Open a BMP that we can re-save
    img_path = os.path.join(resource_path, "surfacetest.bmp")
    sf = sdl2ext.load_bmp(img_path)
    assert isinstance(sf, surf.SDL_Surface)

    # Try saving the BMP to a new folder and re-loading it
    outpath = os.path.join(str(tmpdir), "save_test.bmp")
    sdl2ext.save_bmp(sf, outpath)
    assert os.path.exists(outpath)
    sf_saved = sdl2ext.load_bmp(outpath)
    assert isinstance(sf_saved, surf.SDL_Surface)
    check_image_contents(sf_saved)

    # Try modifying/overwriting the existing BMP
    sdl2ext.fill(sf, (0, 255, 0, 255))
    sdl2ext.save_bmp(sf, outpath, overwrite=True)
    sf_saved2 = sdl2ext.load_bmp(outpath)
    assert isinstance(sf_saved2, surf.SDL_Surface)
    with pytest.raises(AssertionError):
        check_image_contents(sf_saved2)

    surf.SDL_FreeSurface(sf)
    surf.SDL_FreeSurface(sf_saved)
    surf.SDL_FreeSurface(sf_saved2)

    # Test existing file exception with overwrite=False
    with pytest.raises(RuntimeError):
        sdl2ext.save_bmp(sf_saved, outpath, overwrite=False)
    
    # Test exception with non-existent save directory
    bad_path = os.path.join(resource_path, "doesnt_exist", "tst.bmp")
    with pytest.raises(IOError):
        sdl2ext.save_bmp(sf_saved, bad_path)


@pytest.mark.skipif(not _HASSDLIMAGE, reason="Requires SDL2_image")
def test_load_img(with_sdl):
    # Test loading all test images, with and without ARGB conversion
    resources = os.listdir(resource_path)
    test_imgs = [f for f in resources if f[:11] == "surfacetest"]
    for img in test_imgs:
        img_path = os.path.join(resource_path, img)
        fmt = img.split(".")[-1]
        if fmt in skip_formats:
            continue

        sf = sdl2ext.load_img(img_path)
        assert isinstance(sf, surf.SDL_Surface)
        assert sf.format.contents.format == pixels.SDL_PIXELFORMAT_ARGB8888
        if fmt not in skip_color_check:
            check_image_contents(sf)
        surf.SDL_FreeSurface(sf)

        sf2 = sdl2ext.load_img(img_path, as_argb=False)
        assert isinstance(sf2, surf.SDL_Surface)
        surf.SDL_FreeSurface(sf2)

    # Test exception on missing file
    bad_path = os.path.join(resource_path, "doesnt_exist.bmp")
    with pytest.raises(IOError):
        sdl2ext.load_img(bad_path)

    # Test exception on bad file type
    bad_type = os.path.join(resource_path, "tuffy.ttf")
    with pytest.raises(sdl2ext.SDLError):
        sdl2ext.load_img(bad_type)


@pytest.mark.skipif(not _HASPIL, reason="Pillow library is not installed")
def test_pillow_to_image(with_sdl):
    # Import an image using Pillow
    from PIL import Image
    img_path = os.path.join(resource_path, "surfacetest.bmp")
    pil_img = Image.open(img_path)

    # Convert the image to an SDL surface and verify it worked
    sf = sdl2ext.pillow_to_surface(pil_img)
    assert isinstance(sf, surf.SDL_Surface)
    check_image_contents(sf)
    surf.SDL_FreeSurface(sf)

    # Try converting a palette image
    palette_img = pil_img.convert("P", palette=Image.WEB)
    sfp = sdl2ext.pillow_to_surface(palette_img)
    pxformat = sfp.format.contents
    assert isinstance(sfp, surf.SDL_Surface)
    check_image_contents(sfp)
    assert pxformat.BytesPerPixel == 4
    surf.SDL_FreeSurface(sfp)

    # Try converting a palette image without ARGB conversion
    sfp2 = sdl2ext.pillow_to_surface(palette_img, False)
    pxformat = sfp2.format.contents
    assert isinstance(sfp2, surf.SDL_Surface)
    assert pxformat.BytesPerPixel == 1
    sdl_palette = pxformat.palette.contents
    pil_palette = palette_img.getpalette()
    assert sdl_palette.colors[0].r == pil_palette[0]
    assert sdl_palette.colors[0].g == pil_palette[1]
    assert sdl_palette.colors[0].b == pil_palette[2]
    surf.SDL_FreeSurface(sfp2)

    # Test loading all supported test images and compare against reference
    resources = os.listdir(resource_path)
    test_imgs = [f for f in resources if f[:11] == "surfacetest"]
    for img in test_imgs:
        fmt = img.split(".")[-1]
        if fmt in ("webp", "xcf", "lbm", "svg"):
            continue
        pil_img = Image.open(os.path.join(resource_path, img))
        sf = sdl2ext.pillow_to_surface(pil_img)
        assert isinstance(sf, surf.SDL_Surface)
        assert sf.format.contents.format == pixels.SDL_PIXELFORMAT_ARGB8888
        if fmt not in skip_color_check:
            check_image_contents(sf)
        surf.SDL_FreeSurface(sf)

@pytest.mark.skipif(not _HASSDLIMAGE, reason="Requires SDL2_image")
def test_load_image(with_sdl):
    resources = os.listdir(resource_path)
    test_imgs = [f for f in resources if f[:11] == "surfacetest"]
    for img in test_imgs:
        img_path = os.path.join(resource_path, img)
        fmt = img.split(".")[-1]
        if fmt in skip_formats:
            continue

        # Try normal loading
        sf = sdl2ext.load_image(img_path)
        assert isinstance(sf, surf.SDL_Surface)

        # Force only PIL
        if _HASPIL and fmt not in ("webp", "xcf", "lbm", "svg"):
            sf = sdl2ext.load_image(img_path, enforce="PIL")
            assert isinstance(sf, surf.SDL_Surface)

        # Force only sdlimage
        sf = sdl2ext.load_image(img_path, enforce="SDL")
        assert isinstance(sf, surf.SDL_Surface)

        # Clean up surface now that we're done with it
        surf.SDL_FreeSurface(sf)



import sys
import pytest
from ctypes import c_int, cast, byref, POINTER
from sdl2 import SDL_TRUE, SDL_FALSE, SDL_GetError
from sdl2 import rect, keyboard, scancode, keycode, video

byteify = lambda x: x.encode("utf-8")

@pytest.fixture
def window(with_sdl):
    flag = video.SDL_WINDOW_INPUT_FOCUS
    w = video.SDL_CreateWindow(b"Test", 10, 40, 32, 24, flag)
    assert SDL_GetError() == b""
    assert isinstance(w.contents, video.SDL_Window)
    yield w
    video.SDL_DestroyWindow(w)


def test_SDL_Keysym():
    keysym = keyboard.SDL_Keysym()
    assert keysym.scancode == 0
    assert keysym.sym == 0
    assert keysym.mod == 0

    keysym = keyboard.SDL_Keysym(1, 2, 3, ord("b"))
    assert keysym.scancode == 1
    assert keysym.sym == 2
    assert keysym.mod == 3

    uval = "\u0220"
    if sys.version_info[0] < 3:
        uval = unichr(8224)
    keysym = keyboard.SDL_Keysym(17, 32, 88, ord(uval))
    assert keysym.scancode == 17
    assert keysym.sym == 32
    assert keysym.mod == 88


def test_SDL_GetKeyboardFocus(window):
    # NOTE: Would need to actually set keyboard focus for window to test,
    # which I can't seem to get working in pytest
    video.SDL_ShowWindow(window)
    focused_win = keyboard.SDL_GetKeyboardFocus()
    if focused_win:
        focused_id = video.SDL_GetWindowID(focused_win)
        assert video.SDL_GetWindowID(window) == focused_id

def test_SDL_GetKeyboardState(with_sdl):
    # NOTE: This function returns a pointer to an array, which we can use
    # directly as an array in Python since we know how long it is from numkeys
    states = (c_int * scancode.SDL_NUM_SCANCODES)()
    numkeys = c_int(0)
    keystates = keyboard.SDL_GetKeyboardState(byref(numkeys))
    assert numkeys.value > 0
    for key in keystates[:numkeys.value]:
        assert key in [0, 1]

def test_SDL_GetSetModState(with_sdl):
    test_states = [
        keycode.KMOD_NUM | keycode.KMOD_CAPS | keycode.KMOD_MODE,
        keycode.KMOD_LCTRL | keycode.KMOD_LSHIFT,
        keycode.KMOD_CAPS,
    ]
    initial = keyboard.SDL_GetModState()
    for state in test_states:
        keyboard.SDL_SetModState(state)
        assert keyboard.SDL_GetModState() == state
    # Try resetting the modifier state to the initial value
    keyboard.SDL_SetModState(initial)
    assert keyboard.SDL_GetModState() == initial

def test_SDL_GetKeyFromScancode(with_sdl):
    # Test with letter keys
    scan_key_offset = 93
    for scan in range(scancode.SDL_SCANCODE_A, scancode.SDL_SCANCODE_Z + 1):
        key = keyboard.SDL_GetKeyFromScancode(scan)
        assert key == scan + scan_key_offset
    # Test with number keys
    scan_key_offset = 19
    for scan in range(scancode.SDL_SCANCODE_1, scancode.SDL_SCANCODE_9 + 1):
        key = keyboard.SDL_GetKeyFromScancode(scan)
        assert key == scan + scan_key_offset

def test_SDL_GetScancodeFromKey(with_sdl):
    # Test with letter keys
    scan_key_offset = 93
    for scan in range(scancode.SDL_SCANCODE_A, scancode.SDL_SCANCODE_Z + 1):
        key = scan + scan_key_offset
        assert keyboard.SDL_GetScancodeFromKey(key) == scan
    # Test with number keys
    scan_key_offset = 19
    for scan in range(scancode.SDL_SCANCODE_1, scancode.SDL_SCANCODE_9 + 1):
        key = scan + scan_key_offset
        assert keyboard.SDL_GetScancodeFromKey(key) == scan
    # Test with unknown key
    key = keyboard.SDL_GetScancodeFromKey(477)
    assert key == scancode.SDL_SCANCODE_UNKNOWN

def test_SDL_GetScancodeName(with_sdl):
    names = range(ord('A'), ord('Z'))
    xoff = 0
    for code in range(scancode.SDL_SCANCODE_A, scancode.SDL_SCANCODE_Z):
        name = keyboard.SDL_GetScancodeName(code)
        assert name == byteify(chr(names[xoff]))
        xoff += 1
    # Test with unknown scancode
    name = keyboard.SDL_GetScancodeName(0)
    assert name == b""

def test_SDL_GetScancodeFromName(with_sdl):
    codes = range(scancode.SDL_SCANCODE_A, scancode.SDL_SCANCODE_Z)
    xoff = 0
    for key in range(ord('a'), ord('z')):
        ch = chr(key)
        code = keyboard.SDL_GetScancodeFromName(byteify(ch))
        assert code == codes[xoff]
        xoff += 1
    # Test with invalid key name
    key = keyboard.SDL_GetScancodeFromName(b"")
    assert key == scancode.SDL_SCANCODE_UNKNOWN

def test_SDL_GetKeyName(with_sdl):
    x = 65  # SDL maps everything against upper-case letters
    for key in range(ord('a'), ord('z')):
        ch = chr(x)
        name = keyboard.SDL_GetKeyName(key)
        assert name == byteify(ch)
        x += 1

def test_SDL_GetKeyFromName(with_sdl):
    # Test with lower-case ASCII characters
    for x in range(26):
        key = keyboard.SDL_GetKeyFromName(byteify(chr(x + 97)))
        assert key == x + 97
    # Test with ASCII numbers
    for x in range(10):
        key = keyboard.SDL_GetKeyFromName(("%d" % x).encode("utf-8"))
        assert key == 48 + x
    # Test handling of unknown values
    val = keyboard.SDL_GetKeyFromName(b"not a key")
    assert val == keycode.SDLK_UNKNOWN

def test_SDL_StartStopTextInput(with_sdl):
    keyboard.SDL_StopTextInput()
    assert SDL_GetError() == b""
    assert keyboard.SDL_IsTextInputActive() == SDL_FALSE
    keyboard.SDL_StartTextInput()
    assert SDL_GetError() == b""
    assert keyboard.SDL_IsTextInputActive() == SDL_TRUE

def test_SDL_SetTextInputRect(with_sdl):
    keyboard.SDL_StartTextInput()
    coords = [(0, 0, 0, 0), (-10, -70, 3, 6), (10, 10, 10, 10)]
    for x, y, w, h in coords:
        r = rect.SDL_Rect(x, y, w, h)
        keyboard.SDL_SetTextInputRect(r)
        assert SDL_GetError() == b""

def test_SDL_HasScreenKeyboardSupport(with_sdl):
    ret = keyboard.SDL_HasScreenKeyboardSupport()
    assert ret in [SDL_TRUE, SDL_FALSE]

def test_SDL_IsScreenKeyboardShown(window):
    ret = keyboard.SDL_IsScreenKeyboardShown(window)
    assert ret in [SDL_TRUE, SDL_FALSE]



from ctypes import Structure, POINTER, c_char_p, c_int
from .dll import _bind
from .stdinc import Uint8

__all__ = [
    # Structs
    "SDL_version",

    # Defines
    "SDL_MAJOR_VERSION", "SDL_MINOR_VERSION", "SDL_PATCHLEVEL",

    # Macro Functions
    "SDL_VERSION", "SDL_VERSIONNUM", "SDL_COMPILEDVERSION",
    "SDL_VERSION_ATLEAST",

    # Functions
    "SDL_GetVersion", "SDL_GetRevision", "SDL_GetRevisionNumber"
]


class SDL_version(Structure):
    _fields_ = [("major", Uint8),
                ("minor", Uint8),
                ("patch", Uint8),
                ]

SDL_MAJOR_VERSION = 2
SDL_MINOR_VERSION = 0
SDL_PATCHLEVEL = 20

def SDL_VERSION(x):
    x.major = SDL_MAJOR_VERSION
    x.minor = SDL_MINOR_VERSION
    x.patch = SDL_PATCHLEVEL

SDL_VERSIONNUM = lambda x, y, z: (x * 1000 + y * 100 + z)
SDL_COMPILEDVERSION = SDL_VERSIONNUM(SDL_MAJOR_VERSION, SDL_MINOR_VERSION, SDL_PATCHLEVEL)
SDL_VERSION_ATLEAST = lambda x, y, z: (SDL_COMPILEDVERSION >= SDL_VERSIONNUM(x, y, z))
SDL_GetVersion = _bind("SDL_GetVersion", [POINTER(SDL_version)])
SDL_GetRevision = _bind("SDL_GetRevision", None, c_char_p)
SDL_GetRevisionNumber = _bind("SDL_GetRevisionNumber", None, c_int) # Deprecated as of 2.0.16, add warning?



import pytest
from sdl2.surface import SDL_CreateRGBSurface, SDL_FreeSurface
from sdl2.rect import SDL_Rect
from sdl2.ext.draw import prepare_color, fill
from sdl2 import ext as sdl2ext

try:
    import numpy
    _HASNUMPY = True
except:
    _HASNUMPY = False


@pytest.mark.skipif(not _HASNUMPY, reason="Numpy not available")
def test_subsurface(with_sdl):
    # Initialize colour and surface/view
    sf = SDL_CreateRGBSurface(0, 10, 10, 32, 0, 0, 0, 0)
    WHITE = prepare_color((255, 255, 255), sf)

    # Test creation of subsurface from parent
    ssf = sdl2ext.subsurface(sf.contents, (0, 0, 5, 5))
    assert ssf.w == 5 and ssf.h == 5

    # Test shared pixels between surface
    fill(ssf, (255, 255, 255))
    view = sdl2ext.pixels3d(sf.contents, False)
    assert all(x == 255 for x in view[0][0][:3])
    assert all(x == 255 for x in view[4][4][:3])
    assert all(x == 0 for x in view[5][5][:3])
    SDL_FreeSurface(ssf)

    # Test creation of subsurface using an SDL_Rect
    subsurf_rect = SDL_Rect(2, 2, 6, 6)
    ssf = sdl2ext.subsurface(sf.contents, subsurf_rect)
    assert ssf.w == 6 and ssf.h == 6
    SDL_FreeSurface(ssf)

    # Test creation of subsurface using a surface pointer
    ssf = sdl2ext.subsurface(sf, (0, 0, 5, 5))
    assert ssf.w == 5 and ssf.h == 5
    SDL_FreeSurface(ssf)

    # Test exceptions on bad input
    with pytest.raises(TypeError):
        sdl2ext.subsurface(WHITE, (0, 0, 5, 5))
    with pytest.raises(TypeError):
        sdl2ext.subsurface(sf, (0, 0, 5))
    with pytest.raises(ValueError):
        sdl2ext.subsurface(sf, (0, 0, 50, 50))

    # Clean up after tests
    SDL_FreeSurface(sf)
